# Copyright (c) 2015, Jan-Piet Mens <jpmens(at)gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: dig
author: Jan-Piet Mens (@jpmens) <jpmens(at)gmail.com>
short_description: Query DNS using the dnspython library
requirements:
  - dnspython (python library, http://www.dnspython.org/)
description:
  - The dig lookup runs queries against DNS servers to retrieve DNS records for a specific name (FQDN - fully qualified domain
    name). It is possible to lookup any DNS record in this manner.
  - There is a couple of different syntaxes that can be used to specify what record should be retrieved, and for which name.
    It is also possible to explicitly specify the DNS server(s) to use for lookups.
  - In its simplest form, the dig lookup plugin can be used to retrieve an IPv4 address (DNS A record) associated with FQDN.
  - In addition to (default) A record, it is also possible to specify a different record type that should be queried. This
    can be done by either passing-in additional parameter of format qtype=TYPE to the dig lookup, or by appending /TYPE to
    the FQDN being queried.
  - If multiple values are associated with the requested record, the results are returned as a comma-separated list. In
    such cases you may want to pass option C(wantlist=true) to the lookup call, or alternatively use C(query) instead of C(lookup),
    which results in the record values being returned as a list over which you can iterate later on.
  - By default, the lookup relies on system-wide configured DNS servers for performing the query. It is also possible to
    explicitly specify DNS servers to query using the @DNS_SERVER_1,DNS_SERVER_2,...,DNS_SERVER_N notation. This needs to
    be passed-in as an additional parameter to the lookup.
options:
  _terms:
    description: Domain(s) to query.
    type: list
    elements: str
  qtype:
    description:
      - Record type to query.
      - V(DLV) has been removed in community.general 6.0.0.
      - V(CAA) has been added in community.general 6.3.0.
    type: str
    default: 'A'
    choices:
      - A
      - ALL
      - AAAA
      - CAA
      - CNAME
      - DNAME
      - DNSKEY
      - DS
      - HINFO
      - LOC
      - MX
      - NAPTR
      - NS
      - NSEC3PARAM
      - PTR
      - RP
      - RRSIG
      - SOA
      - SPF
      - SRV
      - SSHFP
      - TLSA
      - TXT
  flat:
    description: If 0 each record is returned as a dictionary, otherwise a string.
    type: int
    default: 1
  retry_servfail:
    description: Retry a nameserver if it returns SERVFAIL.
    default: false
    type: bool
    version_added: 3.6.0
  fail_on_error:
    description:
      - Abort execution on lookup errors.
      - The default for this option is likely to change to V(true) in the future. The current default, V(false), is used for
        backwards compatibility, and results in empty strings or the string V(NXDOMAIN) in the result in case of errors.
    default: false
    type: bool
    version_added: 5.4.0
  real_empty:
    description:
      - Return empty result without empty strings, and return empty list instead of V(NXDOMAIN).
      - The default for this option is likely to change to V(true) in the future.
      - This option is forced to V(true) if multiple domains to be queried are specified.
    default: false
    type: bool
    version_added: 6.0.0
  class:
    description:
      - Class.
    type: str
    default: 'IN'
  tcp:
    description: Use TCP to lookup DNS records.
    default: false
    type: bool
    version_added: 7.5.0
  port:
    description: Use port as target port when looking up DNS records.
    default: 53
    type: int
    version_added: 9.5.0
notes:
  - V(ALL) is not a record in itself, merely the listed fields are available for any record results you retrieve in the form
    of a dictionary.
  - While the plugin supports anything which C(dnspython) supports out of the box, only a subset can be converted into a dictionary.
  - If you need to obtain the AAAA record (IPv6 address), you must specify the record type explicitly. Syntax for specifying
    the record type is shown in the examples below.
  - The trailing dot in most of the examples listed is purely optional, but is specified for completeness/correctness sake.
"""

EXAMPLES = r"""
- name: Simple A record (IPV4 address) lookup for example.com
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.dig', 'example.com.')}}"

- name: "The TXT record for example.org."
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.dig', 'example.org.', qtype='TXT') }}"

- name: "The TXT record for example.org, alternative syntax."
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.dig', 'example.org./TXT') }}"

- name: use in a loop
  ansible.builtin.debug:
    msg: "MX record for gmail.com {{ item }}"
  with_items: "{{ lookup('community.general.dig', 'gmail.com./MX', wantlist=true) }}"

- name: Lookup multiple names at once
  ansible.builtin.debug:
    msg: "A record found {{ item }}"
  loop: "{{ query('community.general.dig', 'example.org.', 'example.com.', 'gmail.com.') }}"

- name: Lookup multiple names at once (from list variable)
  ansible.builtin.debug:
    msg: "A record found {{ item }}"
  loop: "{{ query('community.general.dig', *hosts) }}"
  vars:
    hosts:
      - example.org.
      - example.com.
      - gmail.com.

- ansible.builtin.debug:
    msg: "Reverse DNS for 192.0.2.5 is {{ lookup('community.general.dig', '192.0.2.5/PTR') }}"
- ansible.builtin.debug:
    msg: "Reverse DNS for 192.0.2.5 is {{ lookup('community.general.dig', '5.2.0.192.in-addr.arpa./PTR') }}"
- ansible.builtin.debug:
    msg: "Reverse DNS for 192.0.2.5 is {{ lookup('community.general.dig', '5.2.0.192.in-addr.arpa.', qtype='PTR') }}"
- ansible.builtin.debug:
    msg: "Querying 198.51.100.23 for IPv4 address for example.com. produces {{ lookup('dig', 'example.com', '@198.51.100.23') }}"

- ansible.builtin.debug:
    msg: "XMPP service for gmail.com. is available at {{ item.target }} on port {{ item.port }}"
  with_items: "{{ lookup('community.general.dig', '_xmpp-server._tcp.gmail.com./SRV', flat=0, wantlist=true) }}"

- name: Retry nameservers that return SERVFAIL
  ansible.builtin.debug:
    msg: "{{ lookup('community.general.dig', 'example.org./A', retry_servfail=true) }}"
"""

RETURN = r"""
_list:
  description:
    - List of composed strings or of dictionaries, with fields depending
      on query type.
  type: list
  elements: raw
  contains:
    ALL:
      description:
        - C(owner), C(ttl), C(type).
    A:
      description:
        - C(address).
    AAAA:
      description:
        - C(address).
    CAA:
      description:
        - C(flags).
        - C(tag).
        - C(value).
      version_added: 6.3.0
    CNAME:
      description:
        - C(target).
    DNAME:
      description:
        - C(target).
    DNSKEY:
      description:
        - C(flags), C(algorithm), C(protocol), C(key).
    DS:
      description:
        - C(algorithm), C(digest_type), C(key_tag), C(digest).
    HINFO:
      description:
        - C(cpu), C(os).
    LOC:
      description:
        - C(latitude), C(longitude), C(altitude), C(size), C(horizontal_precision), C(vertical_precision).
    MX:
      description:
        - C(preference), C(exchange).
    NAPTR:
      description:
        - C(order), C(preference), C(flags), C(service), C(regexp), C(replacement).
    NS:
      description:
        - C(target).
    NSEC3PARAM:
      description:
        - C(algorithm), C(flags), C(iterations), C(salt).
    PTR:
      description:
        - C(target).
    RP:
      description:
        - C(mbox), C(txt).
    SOA:
      description:
        - C(mname), C(rname), C(serial), C(refresh), C(retry), C(expire), C(minimum).
    SPF:
      description:
        - C(strings).
    SRV:
      description:
        - C(priority), C(weight), C(port), C(target).
    SSHFP:
      description:
        - C(algorithm), C(fp_type), C(fingerprint).
    TLSA:
      description:
        - C(usage), C(selector), C(mtype), C(cert).
    TXT:
      description:
        - C(strings).
"""

import socket

from ansible.errors import AnsibleError
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms

try:
    import dns.exception
    import dns.name
    import dns.rdataclass
    import dns.resolver
    import dns.reversename
    from dns.rdatatype import (
        AAAA,
        CAA,
        CNAME,
        DNAME,
        DNSKEY,
        DS,
        HINFO,
        LOC,
        MX,
        NAPTR,
        NS,
        NSEC3PARAM,
        PTR,
        RP,
        SOA,
        SPF,
        SRV,
        SSHFP,
        TLSA,
        TXT,
        A,
    )

    HAVE_DNS = True
except ImportError:
    HAVE_DNS = False


display = Display()


def make_rdata_dict(rdata):
    """While the 'dig' lookup plugin supports anything which dnspython supports
    out of the box, the following supported_types list describes which
    DNS query types we can convert to a dict.

    Note: adding support for RRSIG is hard work. :)
    """
    supported_types = {
        A: ["address"],
        AAAA: ["address"],
        CAA: ["flags", "tag", "value"],
        CNAME: ["target"],
        DNAME: ["target"],
        DNSKEY: ["flags", "algorithm", "protocol", "key"],
        DS: ["algorithm", "digest_type", "key_tag", "digest"],
        HINFO: ["cpu", "os"],
        LOC: ["latitude", "longitude", "altitude", "size", "horizontal_precision", "vertical_precision"],
        MX: ["preference", "exchange"],
        NAPTR: ["order", "preference", "flags", "service", "regexp", "replacement"],
        NS: ["target"],
        NSEC3PARAM: ["algorithm", "flags", "iterations", "salt"],
        PTR: ["target"],
        RP: ["mbox", "txt"],
        # RRSIG: ['type_covered', 'algorithm', 'labels', 'original_ttl', 'expiration', 'inception', 'key_tag', 'signer', 'signature'],
        SOA: ["mname", "rname", "serial", "refresh", "retry", "expire", "minimum"],
        SPF: ["strings"],
        SRV: ["priority", "weight", "port", "target"],
        SSHFP: ["algorithm", "fp_type", "fingerprint"],
        TLSA: ["usage", "selector", "mtype", "cert"],
        TXT: ["strings"],
    }

    rd = {}

    if rdata.rdtype in supported_types:
        fields = supported_types[rdata.rdtype]
        for f in fields:
            val = rdata.__getattribute__(f)

            if isinstance(val, dns.name.Name):
                val = dns.name.Name.to_text(val)

            if rdata.rdtype == DS and f == "digest":
                val = dns.rdata._hexify(rdata.digest).replace(" ", "")
            if rdata.rdtype == DNSKEY and f == "algorithm":
                val = int(val)
            if rdata.rdtype == DNSKEY and f == "key":
                val = dns.rdata._base64ify(rdata.key).replace(" ", "")
            if rdata.rdtype == NSEC3PARAM and f == "salt":
                val = dns.rdata._hexify(rdata.salt).replace(" ", "")
            if rdata.rdtype == SSHFP and f == "fingerprint":
                val = dns.rdata._hexify(rdata.fingerprint).replace(" ", "")
            if rdata.rdtype == TLSA and f == "cert":
                val = dns.rdata._hexify(rdata.cert).replace(" ", "")

            rd[f] = val

    return rd


# ==============================================================
# dig: Lookup DNS records
#
# --------------------------------------------------------------


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        """
        terms contains a string with things to `dig' for. We support the
        following formats:
            example.com                                     # A record
            example.com  qtype=A                            # same
            example.com/TXT                                 # specific qtype
            example.com  qtype=txt                          # same
            192.0.2.23/PTR                                 # reverse PTR
              ^^ shortcut for 23.2.0.192.in-addr.arpa/PTR
            example.net/AAAA  @nameserver                   # query specified server
                               ^^^ can be comma-sep list of names/addresses

            ... flat=0                                      # returns a dict; default is 1 == string
        """
        if HAVE_DNS is False:
            raise AnsibleError("The dig lookup requires the python 'dnspython' library and it is not installed")

        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)

        # Create Resolver object so that we can set NS if necessary
        myres = dns.resolver.Resolver(configure=True)
        edns_size = 4096
        myres.use_edns(0, ednsflags=dns.flags.DO, payload=edns_size)

        domains = []
        nameservers = []
        qtype = self.get_option("qtype")
        flat = self.get_option("flat")
        fail_on_error = self.get_option("fail_on_error")
        real_empty = self.get_option("real_empty")
        tcp = self.get_option("tcp")
        port = self.get_option("port")
        try:
            rdclass = dns.rdataclass.from_text(self.get_option("class"))
        except Exception as e:
            raise AnsibleError(f"dns lookup illegal CLASS: {e}") from e
        myres.retry_servfail = self.get_option("retry_servfail")

        for t in terms:
            if t.startswith("@"):  # e.g. "@10.0.1.2,192.0.2.1" is ok.
                nsset = t[1:].split(",")
                for ns in nsset:
                    # Check if we have a valid IP address. If so, use that, otherwise
                    # try to resolve name to address using system's resolver. If that
                    # fails we bail out.
                    try:
                        socket.inet_aton(ns)
                        nameservers.append(ns)
                    except Exception:
                        try:
                            nsaddr = dns.resolver.query(ns)[0].address
                            nameservers.append(nsaddr)
                        except Exception as e:
                            raise AnsibleError(f"dns lookup NS: {e}") from e
                continue
            if "=" in t:
                try:
                    opt, arg = t.split("=", 1)
                except Exception:
                    pass

                if opt == "qtype":
                    qtype = arg.upper()
                elif opt == "flat":
                    flat = int(arg)
                elif opt == "class":
                    try:
                        rdclass = dns.rdataclass.from_text(arg)
                    except Exception as e:
                        raise AnsibleError(f"dns lookup illegal CLASS: {e}") from e
                elif opt == "retry_servfail":
                    myres.retry_servfail = boolean(arg)
                elif opt == "fail_on_error":
                    fail_on_error = boolean(arg)
                elif opt == "real_empty":
                    real_empty = boolean(arg)
                elif opt == "tcp":
                    tcp = boolean(arg)

                continue

            if "/" in t:
                try:
                    domain, qtype = t.split("/")
                    domains.append(domain)
                except Exception:
                    domains.append(t)
            else:
                domains.append(t)

        # print "--- domain = {domain} qtype={qtype} rdclass={rdclass}"

        if port:
            myres.port = port
        if len(nameservers) > 0:
            myres.nameservers = nameservers

        if qtype.upper() == "PTR":
            reversed_domains = []
            for domain in domains:
                try:
                    n = dns.reversename.from_address(domain)
                    reversed_domains.append(n.to_text())
                except dns.exception.SyntaxError:
                    pass
                except Exception as e:
                    raise AnsibleError(f"dns.reversename unhandled exception {e}") from e
            domains = reversed_domains

        if len(domains) > 1:
            real_empty = True

        ret = []

        for domain in domains:
            try:
                answers = myres.query(domain, qtype, rdclass=rdclass, tcp=tcp)
                for rdata in answers:
                    s = rdata.to_text()
                    if qtype.upper() == "TXT":
                        s = s[1:-1]  # Strip outside quotes on TXT rdata

                    if flat:
                        ret.append(s)
                    else:
                        try:
                            rd = make_rdata_dict(rdata)
                            rd["owner"] = answers.canonical_name.to_text()
                            rd["type"] = dns.rdatatype.to_text(rdata.rdtype)
                            rd["ttl"] = answers.rrset.ttl
                            rd["class"] = dns.rdataclass.to_text(rdata.rdclass)

                            ret.append(rd)
                        except Exception as err:
                            if fail_on_error:
                                raise AnsibleError(f"Lookup failed: {err}") from err
                            ret.append(str(err))

            except dns.resolver.NXDOMAIN as err:
                if fail_on_error:
                    raise AnsibleError(f"Lookup failed: {err}") from err
                if not real_empty:
                    ret.append("NXDOMAIN")
            except (dns.resolver.NoAnswer, dns.resolver.Timeout, dns.resolver.NoNameservers) as err:
                if fail_on_error:
                    raise AnsibleError(f"Lookup failed: {err}") from err
                if not real_empty:
                    ret.append("")
            except dns.exception.DNSException as err:
                raise AnsibleError(f"dns.resolver unhandled exception {err}") from err

        return ret



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Sudhakar Shet Kudtarkar (@kudtarkar1)
# Copyright: (c) 2020, Shreyas Srish <ssrish@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}


DOCUMENTATION = r"""
---
module: aci_l2out_extepg
short_description: Manage External Network Instance (L2Out External EPG) objects (l2ext:InstP).
description:
- Manage External Network Instance (L2Out External EPG) objects on ACI fabrics.
options:
  tenant:
    description:
    - Name of existing tenant.
    type: str
  l2out:
    description:
    - Name of the l2out.
    type: str
  extepg:
    description:
    - Name of the external end point group.
    type: str
    aliases: [ external_epg, extepg_name, name ]
  description:
    description:
    - Description for the l2out.
    type: str
  preferred_group:
    description:
    - This depicts whether this External EPG is part of the Preferred Group and can communicate without contracts.
    - This is convenient for migration scenarios, or when ACI is used for network automation but not for policy.
    - The APIC defaults to C(false) when unset during creation.
    type: bool
  qos_class:
    description:
    - The bandwidth level for Quality of service.
    type: str
    choices: [ level1, level2, level3, level4, level5, level6, Unspecified ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(tenant) and C(l2out) must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant) and M(cisco.aci.aci_l2out) modules can be used for this.
seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(l2ext:InstP).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Sudhakar Shet Kudtarkar (@kudtarkar1)
- Shreyas Srish (@shrsr)
"""

EXAMPLES = r"""
- name: Add a new L2 external end point group
  cisco.aci.aci_l2out_extepg:
    host: apic
    username: admin
    password: SomeSecretePassword
    tenant: Auto-Demo
    l2out: l2out
    extepg: NewExt
    description: external epg
    preferred_group: false
    state: present
    delegate_to: localhost

- name: Remove an L2 external end point group
  cisco.aci.aci_l2out_extepg:
    host: apic
    username: admin
    password: SomeSecretePassword
    tenant: Auto-Demo
    l2out: l2out
    extepg: NewExt
    state: absent
    delegate_to: localhost

- name: Query the L2 external end point group
  cisco.aci.aci_l2out_extepg:
    host: apic
    username: admin
    password: SomeSecretePassword
    tenant: Auto-Demo
    l2out: l2out
    extepg: NewExt
    state: query
    delegate_to: localhost
    register: query_result

- name: Query all L2 external end point groups in a tenant
  cisco.aci.aci_l2out_extepg:
    host: apic
    username: admin
    password: SomeSecretePassword
    tenant: Auto-Demo
    state: query
    delegate_to: localhost
    register: query_result
"""

RETURN = r"""
   current:
     description: The existing configuration from the APIC after the module has finished
     returned: success
     type: list
     sample:
       [
           {
               "fvTenant": {
                   "attributes": {
                       "descr": "Production environment",
                       "dn": "uni/tn-production",
                       "name": "production",
                       "nameAlias": "",
                       "ownerKey": "",
                       "ownerTag": ""
                   }
               }
           }
       ]
   error:
     description: The error information as returned from the APIC
     returned: failure
     type: dict
     sample:
       {
           "code": "122",
           "text": "unknown managed object class foo"
       }
   raw:
     description: The raw output returned by the APIC REST API (xml or json)
     returned: parse error
     type: str
     sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class "/></imdata>'
   sent:
     description: The actual/minimal configuration pushed to the APIC
     returned: info
     type: list
     sample:
       {
           "fvTenant": {
               "attributes": {
                   "descr": "Production environment"
               }
           }
       }
   previous:
     description: The original configuration from the APIC before the module has started
     returned: info
     type: list
     sample:
       [
           {
               "fvTenant": {
                   "attributes": {
                       "descr": "Production",
                       "dn": "uni/tn-production",
                       "name": "production",
                       "nameAlias": "",
                       "ownerKey": "",
                       "ownerTag": ""
                   }
               }
           }
       ]
   proposed:
     description: The assembled configuration from the user-provided parameters
     returned: info
     type: dict
     sample:
       {
           "fvTenant": {
               "attributes": {
                   "descr": "Production environment",
                   "name": "production"
               }
           }
       }
   filter_string:
     description: The filter string used for the request
     returned: failure or debug
     type: str
     sample: ?rsp-prop-include=config-only
   method:
     description: The HTTP method used for the request to the APIC
     returned: failure or debug
     type: str
     sample: POST
   response:
     description: The HTTP response from the APIC
     returned: failure or debug
     type: str
     sample: OK (30 bytes)
   status:
     description: The HTTP status from the APIC
     returned: failure or debug
     type: int
     sample: 200
   url:
     description: The HTTP url used for the request to the APIC
     returned: failure or debug
     type: str
     sample: https://10.11.12.13/api/mo/uni/tn-production.json
   """

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        l2out=dict(type="str"),
        description=dict(type="str"),
        extepg=dict(type="str", aliases=["external_epg", "extepg_name", "name"]),
        preferred_group=dict(type="bool"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        tenant=dict(type="str"),
        qos_class=dict(type="str", choices=["level1", "level2", "level3", "level4", "level5", "level6", "Unspecified"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["l2out", "tenant", "extepg"]],
            ["state", "present", ["l2out", "tenant", "extepg"]],
        ],
    )

    aci = ACIModule(module)

    l2out = module.params.get("l2out")
    description = module.params.get("description")
    preferred_group = aci.boolean(module.params.get("preferred_group"), "include", "exclude")
    state = module.params.get("state")
    tenant = module.params.get("tenant")
    extepg = module.params.get("extepg")
    qos_class = module.params.get("qos_class")

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="l2extOut",
            aci_rn="l2out-{0}".format(l2out),
            module_object=l2out,
            target_filter={"name": l2out},
        ),
        subclass_2=dict(
            aci_class="l2extInstP",
            aci_rn="instP-{0}".format(extepg),
            module_object=extepg,
            target_filter={"name": extepg},
        ),
    )

    aci.get_existing()

    if state == "present":
        config = dict(name=extepg, descr=description, dn="uni/tn-{0}/l2out-{1}/instP-{2}".format(tenant, l2out, extepg), prefGrMemb=preferred_group)
        if qos_class:
            config.update(prio=qos_class)
        aci.payload(
            class_config=config,
            aci_class="l2extInstP",
        )

        aci.get_diff(aci_class="l2extInstP")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# Copyright 2022 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import os
import tarfile


class ImageArchiveManifestSummary:
    """
    Represents data extracted from a manifest.json found in the tar archive output of the
    "docker image save some:tag > some.tar" command.
    """

    def __init__(self, image_id: str, repo_tags: list[str]) -> None:
        """
        :param image_id:  File name portion of Config entry, e.g. abcde12345 from abcde12345.json
        :param repo_tags  Docker image names, e.g. ["hello-world:latest"]
        """

        self.image_id = image_id
        self.repo_tags = repo_tags


class ImageArchiveInvalidException(Exception):
    pass


def api_image_id(archive_image_id: str) -> str:
    """
    Accepts an image hash in the format stored in manifest.json, and returns an equivalent identifier
    that represents the same image hash, but in the format presented by the Docker Engine API.

    :param archive_image_id: plain image hash
    :returns: Prefixed hash used by REST api
    """

    return f"sha256:{archive_image_id}"


def load_archived_image_manifest(
    archive_path: str,
) -> list[ImageArchiveManifestSummary] | None:
    """
    Attempts to get image IDs and image names from metadata stored in the image
    archive tar file.

    The tar should contain a file "manifest.json" with an array with one or more entries,
    and every entry should have a Config field with the image ID in its file name, as
    well as a RepoTags list, which typically has only one entry.

    :raises:
        ImageArchiveInvalidException: A file already exists at archive_path, but could not extract an image ID from it.

    :param archive_path: Tar file to read
    :return: None, if no file at archive_path, or a list of ImageArchiveManifestSummary objects.
    """

    try:
        # FileNotFoundError does not exist in Python 2
        if not os.path.isfile(archive_path):
            return None

        with tarfile.open(archive_path, "r") as tf:
            try:
                try:
                    reader = tf.extractfile("manifest.json")
                    if reader is None:
                        raise ImageArchiveInvalidException(
                            "Failed to read manifest.json"
                        )
                    with reader as ef:
                        manifest = json.load(ef)
                except ImageArchiveInvalidException:
                    raise
                except Exception as exc:
                    raise ImageArchiveInvalidException(
                        f"Failed to decode and deserialize manifest.json: {exc}"
                    ) from exc

                if len(manifest) == 0:
                    raise ImageArchiveInvalidException(
                        "Expected to have at least one entry in manifest.json but found none"
                    )

                result = []
                for index, meta in enumerate(manifest):
                    try:
                        config_file = meta["Config"]
                    except KeyError as exc:
                        raise ImageArchiveInvalidException(
                            f"Failed to get Config entry from {index + 1}th manifest in manifest.json: {exc}"
                        ) from exc

                    # Extracts hash without 'sha256:' prefix
                    try:
                        # Strip off .json filename extension, leaving just the hash.
                        image_id = os.path.splitext(config_file)[0]
                    except Exception as exc:
                        raise ImageArchiveInvalidException(
                            f"Failed to extract image id from config file name {config_file}: {exc}"
                        ) from exc

                    for prefix in ("blobs/sha256/",):  # Moby 25.0.0, Docker API 1.44
                        if image_id.startswith(prefix):
                            image_id = image_id[len(prefix) :]

                    try:
                        repo_tags = meta["RepoTags"]
                    except KeyError as exc:
                        raise ImageArchiveInvalidException(
                            f"Failed to get RepoTags entry from {index + 1}th manifest in manifest.json: {exc}"
                        ) from exc

                    result.append(
                        ImageArchiveManifestSummary(
                            image_id=image_id, repo_tags=repo_tags
                        )
                    )
                return result

            except ImageArchiveInvalidException:
                raise
            except Exception as exc:
                raise ImageArchiveInvalidException(
                    f"Failed to extract manifest.json from tar file {archive_path}: {exc}"
                ) from exc

    except ImageArchiveInvalidException:
        raise
    except Exception as exc:
        raise ImageArchiveInvalidException(
            f"Failed to open tar file {archive_path}: {exc}"
        ) from exc


def archived_image_manifest(archive_path: str) -> ImageArchiveManifestSummary | None:
    """
    Attempts to get Image.Id and image name from metadata stored in the image
    archive tar file.

    The tar should contain a file "manifest.json" with an array with a single entry,
    and the entry should have a Config field with the image ID in its file name, as
    well as a RepoTags list, which typically has only one entry.

    :raises:
        ImageArchiveInvalidException: A file already exists at archive_path, but could not extract an image ID from it.

    :param archive_path: Tar file to read
    :return: None, if no file at archive_path, or the extracted image ID, which will not have a sha256: prefix.
    """

    results = load_archived_image_manifest(archive_path)
    if results is None:
        return None
    if len(results) == 1:
        return results[0]
    raise ImageArchiveInvalidException(
        f"Expected to have one entry in manifest.json but found {len(results)}"
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The module file for ios_vrf_address_family
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: ios_vrf_address_family
short_description: Resource module to configure VRF definitions.
description: This module provides declarative management of VRF definitions on Cisco IOS.
version_added: 7.0.0
author: Ruchi Pakhle (@Ruchip16)
notes:
  - Tested against Cisco IOSXE version 17.3 on CML.
  - This module works with connection C(network_cli).
    See U(https://docs.ansible.com/ansible/latest/network/user_guide/platform_ios.html)
options:
  config:
    description: A list of device configurations for VRF.
    type: list
    elements: dict
    suboptions:
      name:
        description: Name of the VRF.
        type: str
        required: true
      address_families:
        description: Enable address family and enter its config mode - AFI/SAFI configuration
        type: list
        elements: dict
        suboptions:
          afi:
            description: Address Family Identifier (AFI)
            type: str
            choices: ['ipv4', 'ipv6']
          safi:
            description: Address Family modifier
            type: str
            choices: ['multicast', 'unicast']
          bgp: &bgp01
            description: Commands pertaining to BGP configuration.
            type: dict
            suboptions:
              next_hop:
                description: Next-hop for the routes of a VRF in the backbone.
                type: dict
                suboptions:
                  loopback:
                    description: Loopback interface for next-hop
                    type: int
          export: &export
            description: VRF export
            type: dict
            suboptions:
              map:
                description: Route-map name
                type: str
              ipv4:
                description: Address family based VRF export
                type: dict
                suboptions:
                  multicast:
                    description: Export prefixes to IPv4 Multicast table
                    type: dict
                    suboptions:
                      prefix:
                        description: Upper limit on export prefixes without hogging memory
                        type: int
                      map:
                        description: Route-map name
                        type: str
                  unicast:
                    description: Export prefixes to IPv4 Unicast table
                    type: dict
                    suboptions:
                      prefix:
                        description: Upper limit on export prefixes without hogging memory
                        type: int
                      map:
                        description: Route-map name
                        type: str
                      allow_evpn:
                        description: Allow EVPN routes into global table
                        type: bool
          import_config: &import01
            description: VRF import
            type: dict
            suboptions:
              map:
                description: Route-map name
                type: str
              ipv4:
                description: Address family based VRF import
                type: dict
                suboptions:
                  multicast:
                    description: Import prefixes from IPv4 Multicast table
                    type: dict
                    suboptions:
                      prefix:
                        description: Upper limit on import prefixes without hogging memory
                        type: int
                      map:
                        description: Route-map name
                        type: str
                  unicast:
                    description: Import prefixes from IPv4 Unicast table
                    type: dict
                    suboptions:
                      limit:
                        description: Upper limit on import prefixes without hogging memory
                        type: int
                      map:
                        description: Route-map based VRF import
                        type: str
                      allow_evpn:
                        description: allow Global->VRF routes into EVPN
                        type: bool
          inter_as_hybrid: &inter_as_hybrid
            description: Inter AS hybrid mode
            type: dict
            suboptions:
              csc:
                description: Carrier Supporting Carrier
                type: dict
                suboptions:
                  next_hop: &next_hop
                    description: Next-hop for the routes of a VRF in the backbone.
                    type: str
              next_hop: *next_hop
          mdt:
            description: Backbone Multicast Distribution Tree
            type: dict
            suboptions:
              auto_discovery:
                description: BGP auto-discovery for MVPN
                type: dict
                suboptions:
                  ingress_replication:
                    description: BGP auto-discovery for Ingress-Replication
                    type: dict
                    suboptions:
                      inter_as:
                        description: Enable Inter-AS BGP auto-discovery
                        type: dict
                        suboptions:
                          mdt_hello_enable: &mdt_hello_enable
                            description: Enable PIM Hellos over MDT interface
                            type: bool
                      mdt_hello_enable: *mdt_hello_enable
                  mldp:
                    description: Use MLDP for MDT auto-discovery.
                    type: bool
                  pim:
                    description: BGP auto-discovery for PIM
                    type: dict
                    suboptions:
                      inter_as:
                        description: Enable Inter-AS BGP auto-discovery
                        type: dict
                        suboptions:
                          mdt_hello_enable: *mdt_hello_enable
                          pim_tlv_announce: &pim_tlv_announce
                            description: Announce PIM TLV for data MDT
                            type: dict
                            suboptions:
                              mdt_hello_enable: *mdt_hello_enable
                      mdt_hello_enable: *mdt_hello_enable
                      pim_tlv_announce: *pim_tlv_announce
                  receiver_site:
                    description: BGP receiver only site for MVPN
                    type: bool
              data:
                description: MDT data trees
                type: dict
                suboptions:
                  ingress_replication:
                    description: Use Ingress-Replication to create the data MDT
                    type: dict
                    suboptions:
                      number:
                        description: Number of data MDT
                        type: int
                      immediate_switch:
                        description: Switch immediately to Data MDT tree
                        type: bool
                      list: &list
                        description: Access-list
                        type: dict
                        suboptions:
                          access_list_number:
                            description: Access-list number
                            type: int
                          access_list_name:
                            description: IP Named Extended Access list
                            type: str
                  list: *list
                  mpls:
                    description: Configure MPLS for data MDT
                    type: dict
                    suboptions:
                      mldp:
                        description: Use MLDP for MPLS data MDT
                        type: int
                  threshold:
                    description: MDT switching threshold
                    type: int
              default:
                description: Default MDT configuration
                type: dict
                suboptions:
                  ingress_replication:
                    description: Use Ingress-Replication for the default MDT
                    type: bool
              direct:
                description: Direct MDT's
                type: bool
              log_reuse:
                description: Event logging for data MDT reuse
                type: bool
              mode:
                description: The type of encapsulation
                type: dict
                suboptions:
                  gre:
                    description: GRE encapsulation
                    type: bool
              mtu:
                description: The MTU
                type: int
              overlay:
                description: MDT overlay Protocol
                type: dict
                suboptions:
                  bgp:
                    description: BGP Overlay signalling
                    type: dict
                    suboptions:
                      shared_tree_prune_delay:
                        description: Delay before shared tree is pruned at C-RP PE
                        type: int
                      source_tree_prune_delay:
                        description: Delay before source tree is pruned at C-S PE
                        type: int
                  use_bgp:
                    description: Use BGP for MDT overlay signaling
                    type: dict
                    suboptions:
                      set:
                        description: Enable the use of BGP for MDT overlay.
                        type: bool
                      spt_only:
                        description: Enable SPT-only ASM mode
                        type: bool
              partitioned:
                description: Partitioned Multicast Distribution Tree
                type: dict
                suboptions:
                  ingress_replication:
                    description: Use Ingress-Replication for the partitioned MDT
                    type: bool
                  mldp:
                    description: Use MLDP for the partitioned MDT.
                    type: dict
                    suboptions:
                      p2mp:
                        description: Enable P2MP for partitioned MLDP MDT.
                        type: bool
              strict_rpf:
                description: Enable strict RPF check
                type: dict
                suboptions:
                  interface:
                    description: Interface based strict RPF check
                    type: bool
          protection:
            description: Configure local repair
            type: dict
            suboptions:
              local_prefixes:
                description: Enable protection for local prefixes
                type: bool
          route_replicate: &route_replicate
            description: Replicate (import) routes from another topology (and another VRF)
            type: dict
            suboptions:
              recursion_policy:
                description: Route replication recursion policy
                type: dict
                suboptions:
                  destination:
                    description: Recurse in destination topology
                    type: bool
              from_config:
                description: Replicate routes from another VRF
                type: dict
                suboptions:
                  multicast:
                    description: Multicast SAFI
                    type: dict
                    suboptions:
                      all: &all
                        description: All routes
                        type: dict
                        suboptions:
                          route_map: &route_map
                            description: Route-map reference
                            type: str
                      bgp: &bgp
                        description: Border Gateway Protocol (BGP)
                        type: dict
                        suboptions:
                          as_number:
                            description: Autonomous System Number
                            type: int
                          route_map: *route_map
                      eigrp: &eigrp
                        description: Enhanced Interior Gateway Routing Protocol (EIGRP)
                        type: dict
                        suboptions:
                          as_number:
                            description: Autonomous System Number
                            type: int
                          route_map: *route_map
                      isis: &isis
                        description: Intermediate System-to-Intermediate System (ISIS)
                        type: dict
                        suboptions:
                          iso_tag:
                            description: ISO routing area tag
                            type: str
                          route_map: *route_map
                      mobile: &mobile
                        description: Mobile routes
                        type: dict
                        suboptions:
                          route_map: *route_map
                      odr: &odr
                        description: On-Demand Stub routes
                        type: dict
                        suboptions:
                          route_map: *route_map
                      ospf: &ospf
                        description: Open Shortest Path First (OSPF)
                        type: dict
                        suboptions:
                          process_id:
                            description: OSPF process ID
                            type: int
                          route_map: *route_map
                      rip: &rip
                        description: Routing Information Protocol (RIP)
                        type: dict
                        suboptions:
                          route_map: *route_map
                      static: &static
                        description: Static routes
                        type: dict
                        suboptions:
                          route_map: *route_map
                      topology: &topology
                        description: Topology name
                        type: dict
                        suboptions:
                          base:
                            description: Base topology
                            type: dict
                            suboptions:
                              all: *all
                              bgp: *bgp
                              eigrp: *eigrp
                              isis: *isis
                              mobile: *mobile
                              odr: *odr
                              ospf: *ospf
                              rip: *rip
                              static: *static
                  unicast:
                    description: Unicast SAFI
                    type: dict
                    suboptions:
                      all: *all
                      bgp: *bgp
                      connected:
                        description: Connected routes
                        type: dict
                        suboptions:
                          route_map: *route_map
                      eigrp: *eigrp
                      isis: *isis
                      mobile: *mobile
                      odr: *odr
                      ospf: *ospf
                      rip: *rip
                      static: *static
                  vrf:
                    description: Specify a source VRF
                    type: dict
                    suboptions:
                      name:
                        description: Source VRF name
                        type: str
                      multicast:
                        description: Multicast SAFI
                        type: dict
                        suboptions:
                          all: *all
                          bgp: *bgp
                          eigrp: *eigrp
                          isis: *isis
                          mobile: *mobile
                          odr: *odr
                          ospf: *ospf
                          rip: *rip
                          static: *static
                          topology: *topology
                      unicast:
                        description: Unicast SAFI
                        type: dict
                        suboptions:
                          all: *all
                          bgp: *bgp
                          connected:
                            description: Connected routes
                            type: dict
                            suboptions:
                              route_map: *route_map
                          eigrp: *eigrp
                          isis: *isis
                          mobile: *mobile
                          odr: *odr
                          ospf: *ospf
                          rip: *rip
                          static: *static
                      global:
                        description: global VRF
                        type: dict
                        suboptions:
                          multicast:
                            description: Multicast SAFI
                            type: dict
                            suboptions:
                              all: *all
                              bgp: *bgp
                              eigrp: *eigrp
                              isis: *isis
                              mobile: *mobile
                              odr: *odr
                              ospf: *ospf
                              rip: *rip
                              static: *static
                              topology: *topology
                          unicast:
                            description: Unicast SAFI
                            type: dict
                            suboptions:
                              all: *all
                              bgp: *bgp
                              connected:
                                description: Connected routes
                                type: dict
                                suboptions:
                                  route_map: *route_map
                              eigrp: *eigrp
                              isis: *isis
                              mobile: *mobile
                              odr: *odr
                              ospf: *ospf
                              rip: *rip
                              static: *static
          route_target: &route_target
            description: Specify Target VPN Extended Communities.
            type: dict
            suboptions:
              export:
                description:
                  - This option is DEPRECATED and is replaced with exports which
                    accepts list of dict input.
                type: str
              import_config:
                description:
                 - This option is DEPRECATED and is replaced with imports which
                   accepts list of dict input.
                type: str
              both:
                description:
                  - This option is DEPRECATED and is replaced with both_options which
                    accepts list of dict input.
                type: str
              exports:
                description: Export Target-VPN community.
                type: list
                elements: dict
                suboptions:
                  rt_value:
                    description: Export Target-VPN community.
                    type: str
                  stitching:
                    description: Export Target-VPN community.
                    type: bool

              imports:
                description: Export Target-VPN community.
                type: list
                elements: dict
                suboptions:
                  rt_value:
                    description: Export Target-VPN community.
                    type: str
                  stitching:
                    description: Export Target-VPN community.
                    type: bool

              both_options:
                description: Both export and import Target-VPN community
                type: list
                elements: dict
                suboptions:
                  rt_value:
                    description: Export Target-VPN community.
                    type: str
                  stitching:
                    description: Export Target-VPN community.
                    type: bool

  running_config:
    description:
      - This option is used only with state I(parsed).
      - The value of this option should be the output received from the IOS device by
        executing the command B(show running-config vrf).
      - The state I(parsed) reads the configuration from C(running_config) option and
        transforms it into Ansible structured data as per the resource module's argspec
        and the value is then returned in the I(parsed) key within the result.
    type: str
  state:
    choices: [parsed, gathered, deleted, merged, replaced, rendered, overridden]
    default: merged
    description:
      - The state the configuration should be left in
      - The states I(rendered), I(gathered) and I(parsed) does not perform any change
        on the device.
      - The state I(rendered) will transform the configuration in C(config) option to
        platform specific CLI commands which will be returned in the I(rendered) key
        within the result. For state I(rendered) active connection to remote host is
        not required.
      - The state I(gathered) will fetch the running configuration from device and transform
        it into structured data in the format as per the resource module argspec and
        the value is returned in the I(gathered) key within the result.
      - The state I(parsed) reads the configuration from C(running_config) option and
        transforms it into JSON format as per the resource module parameters and the
        value is returned in the I(parsed) key within the result. The value of C(running_config)
        option should be the same format as the output of command I(show running-config vrf).
        connection to remote host is not required.
    type: str
"""

EXAMPLES = """
# Using merged
#
# Before state:
# -------------
#
# RP/0/0/CPU0:ios#show running-config | section ^vrf
# vrf defnition test
#

- name: Merge provided configuration with device configuration
  cisco.ios.ios_vrf_address_family:
    config:
      - name: test1
        address_families:
          - afi: "ipv4"
            safi: "unicast"
            bgp:
              next_hop:
                loopback: 40
            export:
              ipv4:
                multicast:
                  map: "single"
                  prefix: 345
                unicast:
                  allow_evpn: true
                  map: "test-map"
                  prefix: 67
              map: "testing-map"
            import_config:
              ipv4:
                multicast:
                  map: "import-map"
                  prefix: 89
                unicast:
                  allow_evpn: true
                  limit: 12
                  map: "ran-map"
              map: "import-map"
    state: merged

# Task Output:
# ------------
#
# before: []
#
# commands:
# - vrf definition test1
# - address-family ipv4 unicast
# - bgp next-hop loopback 40
# - export map "testing-map"
# - export ipv4 multicast 345 map "single"
# - export ipv4 unicast 67 map "test-map" allow-evpn
# - import map "import-map"
# - import ipv4 multicast 89 map "import-map"
# - import ipv4 unicast 12 map "ran-map" allow-evpn
#
# after:
# - name: test1
#   address_families:
#     - afi: "ipv4"
#       safi: "unicast"
#       bgp:
#         next_hop:
#           loopback: 40
#       export:
#         ipv4:
#           multicast:
#             map: "single"
#             prefix: 345
#           unicast:
#             allow_evpn: true
#             map: "test-map"
#             prefix: 67
#         map: "testing-map"
#       import_config:
#         ipv4:
#           multicast:
#             map: "import-map"
#             prefix: 89
#           unicast:
#             allow_evpn: true
#             limit: 12
#             map: "ran-map"
#       map: "import-map"
#
# After state:
# ------------
#
# RP/0/0/CPU0:ios#show running-config | section ^vrf
# vrf definition test1
#  address-family ipv4 unicast
#   bgp next-hop loopback 40
#   export map "testing-map"
#   export ipv4 multicast 345 map "single"
#   export ipv4 unicast 67 map "test-map" allow-evpn
#   import map "import-map"
#   import ipv4 multicast 89 map "import-map"
#   import ipv4 unicast 12 map "ran-map" allow-evpn

# Using replaced
#
# Before state:
# -------------
#
# RP/0/0/CPU0:ios#show running-config | section ^vrf
# vrf definition test1
#  address-family ipv4 unicast
#   bgp next-hop loopback 40
#   export map "testing-map"
#   export ipv4 multicast 345 map "single"
#   export ipv4 unicast 67 map "test-map" allow-evpn
#   import map "import-map"
#   import ipv4 multicast 89 map "import-map"
#   import ipv4 unicast 12 map "ran-map" allow-evpn

- name: Replace the provided configuration with the existing running configuration
  cisco.ios.ios_vrf_address_family:
    config:
      - name: test1
        address_families:
          - afi: "ipv4"
            safi: "unicast"
            bgp:
              next_hop:
                loopback: 40
            export:
              ipv4:
                multicast:
                  map: "single"
                  prefix: 345
                unicast:
                  allow_evpn: true
                  map: "test-map"
                  prefix: 67
              map: "testing-map"
            import_config:
              ipv4:
                multicast:
                  map: "import-map"
                  prefix: 89
                unicast:
                  allow_evpn: true
                  limit: 12
                  map: "ran-map"
              map: "import-map"
    state: replaced

# Task Output:
# ------------
#
# before:
# - name: test1
#   address_families:
#     - afi: "ipv4"
#       safi: "unicast"
#       bgp:
#         next_hop:
#           loopback: 40
#       export:
#         ipv4:
#           multicast:
#             map: "single"
#             prefix: 345
#           unicast:
#             allow_evpn: true
#             map: "test-map"
#             prefix: 67
#         map: "testing-map"
#       import_config:
#         ipv4:
#           multicast:
#             map: "import-map"
#             prefix: 89
#           unicast:
#             allow_evpn: true
#             limit: 12
#             map: "ran-map"
#       map: "import-map"
#
# commands:
# - vrf definition test1
# - address-family ipv4 unicast
# - bgp next-hop loopback 40
# - export map "testing-map"
# - export ipv4 multicast 345 map "single"
# - export ipv4 unicast 67 map "test-map" allow-evpn
# - import map "import-map"
# - import ipv4 multicast 89 map "import-map"
# - import ipv4 unicast 12 map "ran-map" allow-evpn
#
# after:
# - name: VRF1
#   address_families:
#     - afi: "ipv4"
#       safi: "unicast"
#       bgp:
#         next_hop:
#           loopback: 23
#       export:
#         ipv4:
#           multicast:
#             map: "single"
#             prefix: 33
#           unicast:
#             allow_evpn: true
#             map: "test-map1"
#             prefix: 7
#         map: "testing-map"
#       import_config:
#         ipv4:
#           multicast:
#             map: "import-map1"
#             prefix: 89
#           unicast:
#             allow_evpn: true
#             limit: 12
#             map: "ran-map"
#       map: "import-map"
#
# After state:
# ------------
#
# RP/0/RP0/CPU0:ios(config)#show running-config vrf
# vrf definition VRF1
#  address-family ipv4 unicast
#   bgp next-hop loopback 23
#   export map "testing-map"
#   export ipv4 multicast 345 map "single"
#   export ipv4 unicast 67 map "test-map1" allow-evpn
#   import map "import-map"
#   import ipv4 multicast 89 map "import-map1"
#   import ipv4 unicast 12 map "ran-map" allow-evpn

# Using overridden
#
# Before state:
# -------------
#
# RP/0/RP0/CPU0:ios(config)#show running-config | section ^vrf
# vrf definition test1
#  address-family ipv4 unicast
#   bgp next-hop loopback 40
#   export map "testing-map"
#   export ipv4 multicast 345 map "single"
#   export ipv4 unicast 67 map "test-map" allow-evpn
#   import map "import-map"
#   import ipv4 multicast 89 map "import-map"
#   import ipv4 unicast 12 map "ran-map" allow-evpn

- name: Override the provided configuration with the existing running configuration
  cisco.ios.ios_vrf_address_family:
    state: overridden
    config:
      - name: VRF7
        address_families:
          - afi: "ipv4"
            safi: "unicast"
            bgp:
              next_hop:
                loopback: 89
            export:
              ipv4:
                multicast:
                  map: "single"
                  prefix: 345
                unicast:
                  allow_evpn: true
                  map: "test-map"
                  prefix: 67
              map: "testing-map"
            import_config:
              ipv4:
                multicast:
                  map: "import-map"
                  prefix: 89
                unicast:
                  allow_evpn: true
                  limit: 12
                  map: "ran-map"
              map: "import-map"
# Task Output:
# ------------
#
# before:
# - name: test1
#   address_families:
#     - afi: "ipv4"
#       safi: "unicast"
#       bgp:
#         next_hop:
#           loopback: 40
#       export:
#         ipv4:
#           multicast:
#             map: "single"
#             prefix: 345
#           unicast:
#             allow_evpn: true
#             map: "test-map"
#             prefix: 67
#         map: "testing-map"
#       import_config:
#         ipv4:
#           multicast:
#             map: "import-map"
#             prefix: 89
#           unicast:
#             allow_evpn: true
#             limit: 12
#             map: "ran-map"
#       map: "import-map"
#
# commands:
# - vrf definition VRF7
# - address-family ipv4 unicast
# - bgp next-hop loopback 89
# - export map "testing-map"
# - export ipv4 multicast 345 map "single"
# - export ipv4 unicast 67 map "test-map" allow-evpn
# - import map "import-map"
# - import ipv4 multicast 89 map "import-map"
# - import ipv4 unicast 12 map "ran-map" allow-evpn

#
# After state:
# -------------
# RP/0/RP0/CPU0:ios(config)#show running-config vrf
# vrf definition VRF4
# vrf definition VRF7
#  address-family ipv4 unicast
#   bgp next-hop loopback 89
#   export map "testing-map"
#   export ipv4 multicast 345 map "single"
#   export ipv4 unicast 67 map "test-map" allow-evpn
#   import map "import-map"
#   import ipv4 multicast 89 map "import-map1"
#   import ipv4 unicast 12 map "ran-map" allow-evpn

# Using deleted
#
# Before state:
# -------------
#
# RP/0/RP0/CPU0:ios(config)#show running-config | section ^vrf
# vrf definition VRF4
# vrf definition VRF6
# address-family ipv4 unicast
#  bgp next-hop loopback 40
#  import map "import-map"
#  import ipv4 multicast 89 map "import-map"
#  import ipv4 unicast 12 map "ran-map" allow-evpn
#  export map "testing-map"
#  export ipv4 multicast 345 map "single"
#  export ipv4 unicast 67 map "test-map" allow-evpn
# vrf definition VRF7

- name: Delete the provided configuration
  cisco.ios.ios_vrf_address_family:
    config:
    state: deleted

# Task Output:
# ------------
#
# before:
# - name: VRF4
# - name: VRF6
#   address_families:
#     - afi: "ipv4"
#       safi: "unicast"
#       bgp:
#         next_hop:
#           loopback: 23
#       import_config:
#         ipv4:
#           multicast:
#             map: "import-map"
#             prefix: 89
#           unicast:
#             map: "ran-map"
#             limit: 12
#             allow_evpn: true
#         map: "import-map"
#       export:
#         ipv4:
#           multicast:
#             map: "single"
#             prefix: 345
#           unicast:
#             map: "test-map"
#             prefix: 67
#             allow_evpn: true
#         map: "testing-map"
# - name: VRF7

# commands:
# - vrf definition VRF4
# - vrf definition VRF6
# - no address-family ipv4 unicast
# - vrf definition VRF7
#
# after:
# - name: VRF4
# - name: VRF6
# - name: VRF7
#
# After state:
# ------------
#
# RP/0/RP0/CPU0:ios(config)#show running-config | section ^vrf
# vrf definition VRF4
# vrf definition VRF6
# vrf definition VRF7

# Using rendered
#
- name: Render provided configuration with device configuration
  cisco.ios.ios_vrf_address_family:
    config:
      - name: test
        address_families:
          - afi: "ipv4"
            safi: "unicast"
            bgp:
              next_hop:
                loopback: 23
            import_config:
              ipv4:
                multicast:
                  map: "import-map"
                  prefix: 89
                unicast:
                  map: "ran-map"
                  limit: 12
                  allow_evpn: true
              map: "import-map"
            export:
              ipv4:
                multicast:
                  map: "single"
                  prefix: 345
                unicast:
                  map: "test-map"
                  prefix: 67
                  allow_evpn: true
              map: "testing-map"
    state: rendered

# Task Output:
# ------------
#
# rendered:
# - vrf definition test
# - address-family ipv4 unicast
# - bgp next-hop loopback 23
# - import map "import-map"
# - import ipv4 multicast 89 map "import-map"
# - import ipv4 unicast 12 map "ran-map" allow-evpn
# - export map "testing-map"
# - export ipv4 multicast 345 map "single"
# - export ipv4 unicast 67 map "test-map" allow-evpn

# Using gathered
#
# Before state:
# -------------
#
# RP/0/RP0/CPU0:ios(config)#show running-config | section ^vrf
# vrf definition test1
# address-family ipv4 unicast
#  bgp next-hop loopback 40
#  import map "import-map"
#  import ipv4 multicast 89 map "import-map"
#  import ipv4 unicast 12 map "ran-map" allow-evpn
#  export map "testing-map"
#  export ipv4 multicast 345 map "single"
#  export ipv4 unicast 67 map "test-map" allow-evpn

- name: Gather existing running configuration
  cisco.ios.ios_vrf_address_family:
    state: gathered

# Task Output:
# ------------
#
# gathered:
# - name: test1
#   address_families:
#    - afi: "ipv4"
#      safi: "unicast"
#      bgp:
#        next_hop:
#          loopback: 40
#      export:
#        ipv4:
#          multicast:
#            map: "single"
#            prefix: 345
#          unicast:
#            allow_evpn: true
#            map: "test-map"
#            prefix: 67
#        map: "testing-map"
#      import_config:
#        ipv4:
#          multicast:
#            map: "import-map"
#            prefix: 89
#          unicast:
#            allow_evpn: true
#            limit: 12
#            map: "ran-map"
#        map: "import-map"

# Using parsed
#
# File: parsed.cfg
# ----------------
#
# vrf definition test
# address-family ipv4 unicast
#  bgp next-hop loopback 23
#  import map "import-map"
#  import ipv4 multicast 89 map "import-map"
#  import ipv4 unicast 12 map "ran-map" allow-evpn
#  export map "testing-map"
#  export ipv4 multicast 345 map "single"
#  export ipv4 unicast 67 map "test-map" allow-evpn

- name: Parse the provided configuration
  cisco.ios.ios_vrf_address_family:
    running_config: "{{ lookup('file', 'parsed.cfg') }}"
    state: parsed

# Task Output:
# ------------
#
# parsed:
#   - address_families:
#     - afi: ipv4
#       bgp:
#         next_hop:
#           loopback: 40
#        export:
#          ipv4:
#            multicast:
#              map: "single"
#              prefix: 345
#            unicast:
#              allow_evpn: true
#              map: "test-map"
#              prefix: 67
#          map: "testing-map"
#        import_config:
#          ipv4:
#            multicast:
#              map: "import-map"
#              prefix: 89
#            unicast:
#              allow_evpn: true
#              limit: 12
#              map: "ran-map"
#       safi: unicast
#     name: test
"""

RETURN = """
before:
  description: The configuration prior to the module execution.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
after:
  description: The resulting configuration after module execution.
  returned: when changed
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
commands:
  description: The set of commands pushed to the remote device.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted)
  type: list
  sample:
    - vrf definition test1
    - address-family ipv4 unicast
    - bgp next-hop loopback 40
    - export map "testing-map"
    - export ipv4 multicast 345 map "testmap"
rendered:
  description: The provided configuration in the task rendered in device-native format (offline).
  returned: when I(state) is C(rendered)
  type: list
  sample:
    - vrf definition test1
    - address-family ipv4 unicast
    - bgp next-hop loopback 40
    - import map "testing-map"
    - export ipv4 multicast 345 map "testmap"
gathered:
  description: Facts about the network resource gathered from the remote device as structured data.
  returned: when I(state) is C(gathered)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
parsed:
  description: The device native config provided in I(running_config) option parsed into structured data as per module argspec.
  returned: when I(state) is C(parsed)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.vrf_address_family.vrf_address_family import (
    Vrf_address_familyArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.config.vrf_address_family.vrf_address_family import (
    Vrf_address_family,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    module = AnsibleModule(
        argument_spec=Vrf_address_familyArgs.argument_spec,
        mutually_exclusive=[["config", "running_config"]],
        required_if=[
            ["state", "merged", ["config"]],
            ["state", "replaced", ["config"]],
            ["state", "overridden", ["config"]],
            ["state", "rendered", ["config"]],
            ["state", "parsed", ["running_config"]],
        ],
        supports_check_mode=True,
    )

    result = Vrf_address_family(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author:
  - Felix Fontein (@felixfontein)
name: docker_api
short_description: Run tasks in docker containers
version_added: 1.1.0
description:
  - Run commands or put/fetch files to an existing docker container.
  - Uses the L(requests library,https://pypi.org/project/requests/) to interact directly with the Docker daemon instead of
    using the Docker CLI. Use the P(community.docker.docker#connection) connection plugin if you want to use the Docker CLI.
notes:
  - Does B(not work with TCP TLS sockets)! This is caused by the inability to send C(close_notify) without closing the connection
    with Python's C(SSLSocket)s. See U(https://github.com/ansible-collections/community.docker/issues/605) for more information.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._docker.var_names
options:
  remote_user:
    type: str
    description:
      - The user to execute as inside the container.
    vars:
      - name: ansible_user
      - name: ansible_docker_user
    ini:
      - section: defaults
        key: remote_user
    env:
      - name: ANSIBLE_REMOTE_USER
    cli:
      - name: user
    keyword:
      - name: remote_user
  remote_addr:
    type: str
    description:
      - The name of the container you want to access.
    default: inventory_hostname
    vars:
      - name: inventory_hostname
      - name: ansible_host
      - name: ansible_docker_host
  container_timeout:
    default: 10
    description:
      - Controls how long we can wait to access reading output from the container once execution started.
    env:
      - name: ANSIBLE_TIMEOUT
      - name: ANSIBLE_DOCKER_TIMEOUT
        version_added: 2.2.0
    ini:
      - key: timeout
        section: defaults
      - key: timeout
        section: docker_connection
        version_added: 2.2.0
    vars:
      - name: ansible_docker_timeout
        version_added: 2.2.0
    cli:
      - name: timeout
    type: integer
  extra_env:
    description:
      - Provide extra environment variables to set when running commands in the Docker container.
      - This option can currently only be provided as Ansible variables due to limitations of ansible-core's configuration
        manager.
    vars:
      - name: ansible_docker_extra_env
    type: dict
    version_added: 3.12.0
  working_dir:
    description:
      - The directory inside the container to run commands in.
      - Requires Docker API version 1.35 or later.
    env:
      - name: ANSIBLE_DOCKER_WORKING_DIR
    ini:
      - key: working_dir
        section: docker_connection
    vars:
      - name: ansible_docker_working_dir
    type: string
    version_added: 3.12.0
  privileged:
    description:
      - Whether commands should be run with extended privileges.
      - B(Note) that this allows command to potentially break out of the container. Use with care!
    env:
      - name: ANSIBLE_DOCKER_PRIVILEGED
    ini:
      - key: privileged
        section: docker_connection
    vars:
      - name: ansible_docker_privileged
    type: boolean
    default: false
    version_added: 3.12.0
"""

import os
import os.path
import typing as t

from ansible.errors import AnsibleConnectionFailure, AnsibleFileNotFound
from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.plugins.connection import ConnectionBase
from ansible.utils.display import Display

from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    APIError,
    DockerException,
    NotFound,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._copy import (
    DockerFileCopyError,
    DockerFileNotFound,
    fetch_file,
    put_file,
)
from ansible_collections.community.docker.plugins.module_utils._version import (
    LooseVersion,
)
from ansible_collections.community.docker.plugins.plugin_utils._common_api import (
    AnsibleDockerClient,
)
from ansible_collections.community.docker.plugins.plugin_utils._socket_handler import (
    DockerSocketHandler,
)

if t.TYPE_CHECKING:
    from collections.abc import Callable

    _T = t.TypeVar("_T")


MIN_DOCKER_API = None


display = Display()


class Connection(ConnectionBase):
    """Local docker based connections"""

    transport = "community.docker.docker_api"
    has_pipelining = True

    def _call_client(
        self,
        f: Callable[[AnsibleDockerClient], _T],
        not_found_can_be_resource: bool = False,
    ) -> _T:
        if self.client is None:
            raise AssertionError("Client must be present")
        remote_addr = self.get_option("remote_addr")
        try:
            return f(self.client)
        except NotFound as e:
            if not_found_can_be_resource:
                raise AnsibleConnectionFailure(
                    f'Could not find container "{remote_addr}" or resource in it ({e})'
                ) from e
            raise AnsibleConnectionFailure(
                f'Could not find container "{remote_addr}" ({e})'
            ) from e
        except APIError as e:
            if e.response is not None and e.response.status_code == 409:
                raise AnsibleConnectionFailure(
                    f'The container "{remote_addr}" has been paused ({e})'
                ) from e
            self.client.fail(
                f'An unexpected Docker error occurred for container "{remote_addr}": {e}'
            )
        except DockerException as e:
            self.client.fail(
                f'An unexpected Docker error occurred for container "{remote_addr}": {e}'
            )
        except RequestException as e:
            self.client.fail(
                f'An unexpected requests error occurred for container "{remote_addr}" when trying to talk to the Docker daemon: {e}'
            )

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        self.client: AnsibleDockerClient | None = None
        self.ids: dict[str | None, tuple[int, int]] = {}

        # Windows uses Powershell modules
        if getattr(self._shell, "_IS_WINDOWS", False):
            self.module_implementation_preferences = (".ps1", ".exe", "")

        self.actual_user: str | None = None

    def _connect(self) -> Connection:
        """Connect to the container. Nothing to do"""
        super()._connect()  # type: ignore[safe-super]
        if not self._connected:
            self.actual_user = self.get_option("remote_user")
            display.vvv(
                f"ESTABLISH DOCKER CONNECTION FOR USER: {self.actual_user or '?'}",
                host=self.get_option("remote_addr"),
            )
            if self.client is None:
                self.client = AnsibleDockerClient(
                    self, min_docker_api_version=MIN_DOCKER_API
                )
            self._connected = True

            if self.actual_user is None and display.verbosity > 2:
                # Since we are not setting the actual_user, look it up so we have it for logging later
                # Only do this if display verbosity is high enough that we'll need the value
                # This saves overhead from calling into docker when we do not need to
                display.vvv("Trying to determine actual user")
                result = self._call_client(
                    lambda client: client.get_json(
                        "/containers/{0}/json", self.get_option("remote_addr")
                    )
                )
                if result.get("Config"):
                    self.actual_user = result["Config"].get("User")
                    if self.actual_user is not None:
                        display.vvv(f"Actual user is '{self.actual_user}'")

        return self

    def exec_command(
        self, cmd: str, in_data: bytes | None = None, sudoable: bool = False
    ) -> tuple[int, bytes, bytes]:
        """Run a command on the docker host"""

        super().exec_command(cmd, in_data=in_data, sudoable=sudoable)  # type: ignore[safe-super]

        if self.client is None:
            raise AssertionError("Client must be present")

        command = [self._play_context.executable, "-c", cmd]

        do_become = self.become and self.become.expect_prompt() and sudoable

        stdin_part = (
            f", with stdin ({len(in_data)} bytes)" if in_data is not None else ""
        )
        become_part = ", with become prompt" if do_become else ""
        display.vvv(
            f"EXEC {to_text(command)}{stdin_part}{become_part}",
            host=self.get_option("remote_addr"),
        )

        need_stdin = bool((in_data is not None) or do_become)

        data = {
            "Container": self.get_option("remote_addr"),
            "User": self.get_option("remote_user") or "",
            "Privileged": self.get_option("privileged"),
            "Tty": False,
            "AttachStdin": need_stdin,
            "AttachStdout": True,
            "AttachStderr": True,
            "Cmd": command,
        }

        if "detachKeys" in self.client._general_configs:
            data["detachKeys"] = self.client._general_configs["detachKeys"]

        if self.get_option("extra_env"):
            data["Env"] = []
            for k, v in self.get_option("extra_env").items():
                for val, what in ((k, "Key"), (v, "Value")):
                    if not isinstance(val, str):
                        raise AnsibleConnectionFailure(
                            f"Non-string {what.lower()} found for extra_env option. Ambiguous env options must be "
                            "wrapped in quotes to avoid them being interpreted when directly specified "
                            "in YAML, or explicitly converted to strings when the option is templated. "
                            f"{what}: {val!r}"
                        )
                data["Env"].append(f"{k}={v}")

        if self.get_option("working_dir") is not None:
            data["WorkingDir"] = self.get_option("working_dir")
            if self.client.docker_api_version < LooseVersion("1.35"):
                raise AnsibleConnectionFailure(
                    "Providing the working directory requires Docker API version 1.35 or newer."
                    f" The Docker daemon the connection is using has API version {self.client.docker_api_version_str}."
                )

        exec_data = self._call_client(
            lambda client: client.post_json_to_json(
                "/containers/{0}/exec", self.get_option("remote_addr"), data=data
            )
        )
        exec_id = exec_data["Id"]

        data = {"Tty": False, "Detach": False}
        if need_stdin:
            exec_socket, response = self._call_client(
                lambda client: client.post_json_to_stream_socket(
                    "/exec/{0}/start", exec_id, data=data
                )
            )
            try:
                with DockerSocketHandler(
                    display, exec_socket, container=self.get_option("remote_addr")
                ) as exec_socket_handler:
                    if do_become:
                        assert self.become is not None

                        become_output = [b""]

                        def append_become_output(stream_id: int, data: bytes) -> None:
                            become_output[0] += data

                        exec_socket_handler.set_block_done_callback(
                            append_become_output
                        )

                        while not self.become.check_success(
                            become_output[0]
                        ) and not self.become.check_password_prompt(become_output[0]):
                            if not exec_socket_handler.select(
                                self.get_option("container_timeout")
                            ):
                                stdout, stderr = exec_socket_handler.consume()
                                raise AnsibleConnectionFailure(
                                    "timeout waiting for privilege escalation password prompt:\n"
                                    + to_text(become_output[0])
                                )

                            if exec_socket_handler.is_eof():
                                raise AnsibleConnectionFailure(
                                    "privilege output closed while waiting for password prompt:\n"
                                    + to_text(become_output[0])
                                )

                        if not self.become.check_success(become_output[0]):
                            become_pass = self.become.get_option(
                                "become_pass", playcontext=self._play_context
                            )
                            exec_socket_handler.write(
                                to_bytes(become_pass, errors="surrogate_or_strict")
                                + b"\n"
                            )

                    if in_data is not None:
                        exec_socket_handler.write(in_data)

                    stdout, stderr = exec_socket_handler.consume()
            finally:
                response.close()
        else:
            stdout, stderr = self._call_client(
                lambda client: client.post_json_to_stream(
                    "/exec/{0}/start",
                    exec_id,
                    stream=False,
                    demux=True,
                    tty=False,
                    data=data,
                )
            )

        result = self._call_client(
            lambda client: client.get_json("/exec/{0}/json", exec_id)
        )

        return result.get("ExitCode") or 0, stdout or b"", stderr or b""

    def _prefix_login_path(self, remote_path: str) -> str:
        """Make sure that we put files into a standard path

        If a path is relative, then we need to choose where to put it.
        ssh chooses $HOME but we are not guaranteed that a home dir will
        exist in any given chroot.  So for now we are choosing "/" instead.
        This also happens to be the former default.

        Can revisit using $HOME instead if it is a problem
        """
        if getattr(self._shell, "_IS_WINDOWS", False):
            import ntpath

            return ntpath.normpath(remote_path)
        if not remote_path.startswith(os.path.sep):
            remote_path = os.path.join(os.path.sep, remote_path)
        return os.path.normpath(remote_path)

    def put_file(self, in_path: str, out_path: str) -> None:
        """Transfer a file from local to docker container"""
        super().put_file(in_path, out_path)  # type: ignore[safe-super]
        display.vvv(f"PUT {in_path} TO {out_path}", host=self.get_option("remote_addr"))

        if self.client is None:
            raise AssertionError("Client must be present")

        out_path = self._prefix_login_path(out_path)

        if self.actual_user not in self.ids:
            dummy, ids, dummy2 = self.exec_command("id -u && id -g")
            remote_addr = self.get_option("remote_addr")
            try:
                b_user_id, b_group_id = ids.splitlines()
                user_id, group_id = int(b_user_id), int(b_group_id)
                self.ids[self.actual_user] = user_id, group_id
                display.vvvv(
                    f'PUT: Determined uid={user_id} and gid={group_id} for user "{self.actual_user}"',
                    host=remote_addr,
                )
            except Exception as e:
                raise AnsibleConnectionFailure(
                    f'Error while determining user and group ID of current user in container "{remote_addr}": {e}\nGot value: {ids!r}'
                ) from e

        user_id, group_id = self.ids[self.actual_user]
        try:
            self._call_client(
                lambda client: put_file(
                    client,
                    container=self.get_option("remote_addr"),
                    in_path=in_path,
                    out_path=out_path,
                    user_id=user_id,
                    group_id=group_id,
                    user_name=self.actual_user,
                    follow_links=True,
                ),
                not_found_can_be_resource=True,
            )
        except DockerFileNotFound as exc:
            raise AnsibleFileNotFound(to_text(exc)) from exc
        except DockerFileCopyError as exc:
            raise AnsibleConnectionFailure(to_text(exc)) from exc

    def fetch_file(self, in_path: str, out_path: str) -> None:
        """Fetch a file from container to local."""
        super().fetch_file(in_path, out_path)  # type: ignore[safe-super]
        display.vvv(
            f"FETCH {in_path} TO {out_path}", host=self.get_option("remote_addr")
        )

        if self.client is None:
            raise AssertionError("Client must be present")

        in_path = self._prefix_login_path(in_path)

        try:
            self._call_client(
                lambda client: fetch_file(
                    client,
                    container=self.get_option("remote_addr"),
                    in_path=in_path,
                    out_path=out_path,
                    follow_links=True,
                    log=lambda msg: display.vvvv(
                        msg, host=self.get_option("remote_addr")
                    ),
                ),
                not_found_can_be_resource=True,
            )
        except DockerFileNotFound as exc:
            raise AnsibleFileNotFound(to_text(exc)) from exc
        except DockerFileCopyError as exc:
            raise AnsibleConnectionFailure(to_text(exc)) from exc

    def close(self) -> None:
        """Terminate the connection. Nothing to do for Docker"""
        super().close()  # type: ignore[safe-super]
        self._connected = False

    def reset(self) -> None:
        self.ids.clear()



# Copyright (c) 2024, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import datetime
import re

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)

UTC = datetime.timezone.utc


def get_now_datetime(*, with_timezone: bool) -> datetime.datetime:
    if with_timezone:
        return datetime.datetime.now(tz=UTC)
    return datetime.datetime.utcnow()


def ensure_utc_timezone(timestamp: datetime.datetime) -> datetime.datetime:
    if timestamp.tzinfo is UTC:
        return timestamp
    if timestamp.tzinfo is None:
        # We assume that naive datetime objects use timezone UTC!
        return timestamp.replace(tzinfo=UTC)
    return timestamp.astimezone(UTC)


def remove_timezone(timestamp: datetime.datetime) -> datetime.datetime:
    # Convert to native datetime object
    if timestamp.tzinfo is None:
        return timestamp
    if timestamp.tzinfo is not UTC:
        timestamp = timestamp.astimezone(UTC)
    return timestamp.replace(tzinfo=None)


def add_or_remove_timezone(
    timestamp: datetime.datetime, *, with_timezone: bool
) -> datetime.datetime:
    return (
        ensure_utc_timezone(timestamp) if with_timezone else remove_timezone(timestamp)
    )


def get_epoch_seconds(timestamp: datetime.datetime) -> float:
    if timestamp.tzinfo is None:
        # timestamp.timestamp() is offset by the local timezone if timestamp has no timezone
        timestamp = ensure_utc_timezone(timestamp)
    return timestamp.timestamp()


def from_epoch_seconds(
    timestamp: int | float, *, with_timezone: bool
) -> datetime.datetime:
    if with_timezone:
        return datetime.datetime.fromtimestamp(timestamp, UTC)
    return datetime.datetime.utcfromtimestamp(timestamp)


def convert_relative_to_datetime(
    relative_time_string: str,
    *,
    with_timezone: bool = False,
    now: datetime.datetime | None = None,
) -> datetime.datetime | None:
    """Get a datetime.datetime or None from a string in the time format described in sshd_config(5)"""

    parsed_result = re.match(
        r"^(?P<prefix>[+-])((?P<weeks>\d+)[wW])?((?P<days>\d+)[dD])?((?P<hours>\d+)[hH])?((?P<minutes>\d+)[mM])?((?P<seconds>\d+)[sS]?)?$",
        relative_time_string,
    )

    if parsed_result is None or len(relative_time_string) == 1:
        # not matched or only a single "+" or "-"
        return None

    offset = datetime.timedelta(0)
    if parsed_result.group("weeks") is not None:
        offset += datetime.timedelta(weeks=int(parsed_result.group("weeks")))
    if parsed_result.group("days") is not None:
        offset += datetime.timedelta(days=int(parsed_result.group("days")))
    if parsed_result.group("hours") is not None:
        offset += datetime.timedelta(hours=int(parsed_result.group("hours")))
    if parsed_result.group("minutes") is not None:
        offset += datetime.timedelta(minutes=int(parsed_result.group("minutes")))
    if parsed_result.group("seconds") is not None:
        offset += datetime.timedelta(seconds=int(parsed_result.group("seconds")))

    if now is None:
        now = get_now_datetime(with_timezone=with_timezone)
    else:
        now = add_or_remove_timezone(now, with_timezone=with_timezone)

    if parsed_result.group("prefix") == "+":
        return now + offset
    return now - offset


def get_relative_time_option(
    input_string: str,
    *,
    input_name: str,
    with_timezone: bool = False,
    now: datetime.datetime | None = None,
) -> datetime.datetime:
    """
    Return an absolute timespec if a relative timespec or an ASN1 formatted
    string is provided.

    The return value will be a datetime object.
    """
    result = to_text(input_string)
    if result is None:
        raise OpenSSLObjectError(
            f'The timespec "{input_string}" for {input_name} is not valid'
        )
    # Relative time
    if result.startswith("+") or result.startswith("-"):
        res = convert_relative_to_datetime(result, with_timezone=with_timezone, now=now)
        if res is None:
            raise OpenSSLObjectError(
                f'The timespec "{input_string}" for {input_name} is invalid'
            )
        return res
    # Absolute time
    for date_fmt, length in [
        (
            "%Y%m%d%H%M%SZ",
            15,
        ),  # this also parses '202401020304Z', but as datetime(2024, 1, 2, 3, 0, 4)
        ("%Y%m%d%H%MZ", 13),
        (
            "%Y%m%d%H%M%S%z",
            14 + 5,
        ),  # this also parses '202401020304+0000', but as datetime(2024, 1, 2, 3, 0, 4, tzinfo=...)
        ("%Y%m%d%H%M%z", 12 + 5),
    ]:
        if len(result) != length:
            continue
        try:
            res = datetime.datetime.strptime(result, date_fmt)
        except ValueError:
            pass
        else:
            return add_or_remove_timezone(res, with_timezone=with_timezone)

    raise OpenSSLObjectError(
        f'The time spec "{input_string}" for {input_name} is invalid'
    )


__all__ = (
    "add_or_remove_timezone",
    "convert_relative_to_datetime",
    "ensure_utc_timezone",
    "from_epoch_seconds",
    "get_epoch_seconds",
    "get_now_datetime",
    "get_relative_time_option",
    "remove_timezone",
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Nirav Katarmal (@nkatarmal-crest) <nirav.katarmal@crestdatasys.com>
# Copyright: (c) 2023, Mabille Florent (@fmabille09) <florent.mabille@smals.be>
# Copyright: (c) 2024, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_anp_epg_domain
short_description: Manage site-local EPG domains in schema template
description:
- Manage site-local EPG domains in schema template on Cisco ACI Multi-Site.
author:
- Nirav Katarmal (@nkatarmal-crest)
- Mabille Florent (@fmabille09)
- Akini Ross (@akinross)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  anp:
    description:
    - The name of the ANP.
    type: str
    required: true
  epg:
    description:
    - The name of the EPG.
    type: str
    required: true
  domain_association_type:
    description:
    - The type of domain to associate.
    type: str
    choices: [ vmmDomain, l3ExtDomain, l2ExtDomain, physicalDomain, fibreChannelDomain ]
  domain_profile:
    description:
    - The domain profile name.
    type: str
  deployment_immediacy:
    description:
    - The deployment immediacy of the domain.
    - C(immediate) means B(Deploy immediate).
    - C(lazy) means B(deploy on demand).
    type: str
    choices: [ immediate, lazy ]
  resolution_immediacy:
    description:
    - Determines when the policies should be resolved and available.
    - Defaults to C(lazy) when unset during creation.
    type: str
    choices: [ immediate, lazy, pre-provision ]
  micro_seg_vlan_type:
    description:
    - Virtual LAN type for microsegmentation. This attribute can only be used with VMM domain association.
    type: str
    choices: [ vlan ]
  micro_seg_vlan:
    description:
    - Virtual LAN for microsegmentation. This attribute can only be used with VMM domain association.
    type: int
  port_encap_vlan_type:
    description:
    - Virtual LAN type for port encap. This attribute can only be used with VMM domain association.
    type: str
    choices: [ vlan ]
  port_encap_vlan:
    description:
    - Virtual LAN type for port encap. This attribute can only be used with VMM domain association.
    type: int
  vlan_encap_mode:
    description:
    - Which VLAN enacap mode to use. This attribute can only be used with VMM domain association.
    type: str
    choices: [ static, dynamic ]
  allow_micro_segmentation:
    description:
    - Specifies microsegmentation is enabled or not. This attribute can only be used with VMM domain association.
    type: bool
  switch_type:
    description:
    - Which switch type to use with this domain association. This attribute can only be used with VMM domain association.
    type: str
  switching_mode:
    description:
    - Which switching mode to use with this domain association. This attribute can only be used with VMM domain association.
    type: str
  enhanced_lagpolicy_name:
    description:
    - EPG enhanced lagpolicy name. This attribute can only be used with VMM domain association.
    type: str
  enhanced_lagpolicy_dn:
    description:
    - Distinguished name of EPG lagpolicy. This attribute can only be used with VMM domain association.
    type: str
  delimiter:
    description:
    - Which delimiter to use with this domain association. This attribute can only be used with VMM domain association.
    type: str
    choices: [ '+', '|', '~', '!', '@', '^', '=' ]
  binding_type:
    description:
    - Which binding_type to use with this domain association. This attribute can only be used with VMM domain association.
    type: str
    default: none
    choices: [ static, dynamic, none, ephemeral ]
  num_ports:
    description:
    - Number of ports for the binding type. This attribute can only be used with VMM domain association.
    type: int
  port_allocation:
    description:
    - Port allocation for the binding type. This attribute can only be used with VMM domain association and binding type in static.
    - Required when O(binding_type=static).
    type: str
    choices: [ elastic, fixed ]
  netflow_pref:
    description:
    - The netflow preference state of the domain association. This attribute can only be used with VMM domain association.
    type: str
    default: disabled
    choices: [ enabled, disabled ]
  allow_promiscuous:
    description:
    - The allow promiscuous state of the domain association. This attribute can only be used with VMM domain association.
    type: str
    default: reject
    choices: [ accept, reject ]
  forged_transmits:
    description:
    - The forged transmits state of the domain association. This attribute can only be used with VMM domain association.
    type: str
    default: reject
    choices: [ accept, reject ]
  mac_changes:
    description:
    - The mac changes state of the domain association. This attribute can only be used with VMM domain association.
    type: str
    default: reject
    choices: [ accept, reject ]
  custom_epg_name:
    description:
    - Which custom_epg_name to use with this domain association. This attribute can only be used with VMM domain association.
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- The ACI MultiSite PATCH API has a deficiency requiring some objects to be referenced by index.
  This can cause silent corruption on concurrent access when changing/removing on object as
  the wrong object may be referenced. This module is affected by this deficiency.
seealso:
- module: cisco.mso.mso_schema_site_anp_epg
- module: cisco.mso.mso_schema_template_anp_epg
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new domain to a site EPG
  cisco.mso.mso_schema_site_anp_epg_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    domain_association_type: vmmDomain
    domain_profile: 'VMware-VMM'
    deployment_immediacy: lazy
    resolution_immediacy: pre-provision
    state: present

- name: Add a new domain to a site EPG with all possible attributes set
  cisco.mso.mso_schema_site_anp_epg_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    domain_association_type: vmmDomain
    domain_profile: 'VMware-VMM'
    deployment_immediacy: immediate
    resolution_immediacy: immediate
    micro_seg_vlan_type: vlan
    micro_seg_vlan: 100
    port_encap_vlan_type: vlan
    port_encap_vlan: 100
    vlan_encap_mode: static
    allow_micro_segmentation: true
    switch_type: default
    switching_mode: native
    enhanced_lagpolicy_name: ansible_lag_name
    enhanced_lagpolicy_dn: ansible_lag_dn
    delimiter: '|'
    binding_type: static
    num_ports: 2
    port_allocation: elastic
    netflow_pref: enabled
    allow_promiscuous: accept
    forged_transmits: accept
    mac_changes: accept
    custom_epg_name: ansible_custom_epg
    state: present

- name: Query a domain associated with a specific site EPG
  cisco.mso.mso_schema_site_anp_epg_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    domain_association_type: vmmDomain
    domain_profile: 'VMware-VMM'
    state: query
  register: query_result

- name: Query all domains associated with a site EPG
  cisco.mso.mso_schema_site_anp_epg_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    state: query
  register: query_result

- name: Remove a domain from a site EPG
  cisco.mso.mso_schema_site_anp_epg_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    domain_association_type: vmmDomain
    domain_profile: 'VMware-VMM'
    deployment_immediacy: lazy
    resolution_immediacy: pre-provision
    state: absent
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        anp=dict(type="str", required=True),
        epg=dict(type="str", required=True),
        domain_association_type=dict(type="str", choices=["vmmDomain", "l3ExtDomain", "l2ExtDomain", "physicalDomain", "fibreChannelDomain"]),
        domain_profile=dict(type="str"),
        deployment_immediacy=dict(type="str", choices=["immediate", "lazy"]),
        resolution_immediacy=dict(type="str", choices=["immediate", "lazy", "pre-provision"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        micro_seg_vlan_type=dict(type="str", choices=["vlan"]),
        micro_seg_vlan=dict(type="int"),
        port_encap_vlan_type=dict(type="str", choices=["vlan"]),
        port_encap_vlan=dict(type="int"),
        vlan_encap_mode=dict(type="str", choices=["static", "dynamic"]),
        allow_micro_segmentation=dict(type="bool"),
        switch_type=dict(type="str"),
        switching_mode=dict(type="str"),
        enhanced_lagpolicy_name=dict(type="str"),
        enhanced_lagpolicy_dn=dict(type="str"),
        binding_type=dict(type="str", default="none", choices=["dynamic", "ephemeral", "none", "static"]),
        port_allocation=dict(type="str", choices=["elastic", "fixed"]),
        num_ports=dict(type="int"),
        netflow_pref=dict(type="str", default="disabled", choices=["enabled", "disabled"]),
        allow_promiscuous=dict(type="str", default="reject", choices=["accept", "reject"]),
        forged_transmits=dict(type="str", default="reject", choices=["accept", "reject"]),
        mac_changes=dict(type="str", default="reject", choices=["accept", "reject"]),
        delimiter=dict(type="str", choices=["+", "|", "~", "!", "@", "^", "="]),
        custom_epg_name=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["domain_association_type", "domain_profile", "deployment_immediacy", "resolution_immediacy"]],
            ["state", "present", ["domain_association_type", "domain_profile", "deployment_immediacy", "resolution_immediacy"]],
            ["binding_type", "static", ["port_allocation"]],
        ],
        required_together=[
            ("micro_seg_vlan_type", "micro_seg_vlan"),
            ("port_encap_vlan_type", "port_encap_vlan"),
            ("enhanced_lagpolicy_name", "enhanced_lagpolicy_dn"),
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    anp = module.params.get("anp")
    epg = module.params.get("epg")
    domain_association_type = module.params.get("domain_association_type")
    domain_profile = module.params.get("domain_profile")
    deployment_immediacy = module.params.get("deployment_immediacy")
    resolution_immediacy = module.params.get("resolution_immediacy")
    state = module.params.get("state")
    micro_seg_vlan_type = module.params.get("micro_seg_vlan_type")
    micro_seg_vlan = module.params.get("micro_seg_vlan")
    port_encap_vlan_type = module.params.get("port_encap_vlan_type")
    port_encap_vlan = module.params.get("port_encap_vlan")
    vlan_encap_mode = module.params.get("vlan_encap_mode")
    allow_micro_segmentation = module.params.get("allow_micro_segmentation")
    switch_type = module.params.get("switch_type")
    switching_mode = module.params.get("switching_mode")
    enhanced_lagpolicy_name = module.params.get("enhanced_lagpolicy_name")
    enhanced_lagpolicy_dn = module.params.get("enhanced_lagpolicy_dn")
    binding_type = module.params.get("binding_type")
    port_allocation = module.params.get("port_allocation")
    num_ports = module.params.get("num_ports")
    netflow_pref = module.params.get("netflow_pref")
    allow_promiscuous = module.params.get("allow_promiscuous")
    forged_transmits = module.params.get("forged_transmits")
    mac_changes = module.params.get("mac_changes")
    delimiter = module.params.get("delimiter")
    custom_epg_name = module.params.get("custom_epg_name")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template
    templates = [t.get("name") for t in schema_obj.get("templates")]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ", ".join(templates)))
    template_idx = templates.index(template)

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if not schema_obj.get("sites"):
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get("siteId"), s.get("templateName")) for s in schema_obj.get("sites")]
    sites_list = [s.get("siteId") + "/" + s.get("templateName") for s in schema_obj.get("sites")]
    if (site_id, template) not in sites:
        mso.fail_json(
            msg="Provided site/siteId/template '{0}/{1}/{2}' does not exist. "
            "Existing siteIds/templates: {3}".format(site, site_id, template, ", ".join(sites_list))
        )

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(site_id, template)

    payload = dict()
    ops = []
    op_path = ""
    domain_path = None

    # Get ANP
    anp_ref = mso.anp_ref(schema_id=schema_id, template=template, anp=anp)
    anps = [a.get("anpRef") for a in schema_obj["sites"][site_idx]["anps"]]
    anps_in_temp = [a.get("name") for a in schema_obj["templates"][template_idx]["anps"]]
    if anp not in anps_in_temp:
        mso.fail_json(msg="Provided anp '{0}' does not exist. Existing anps: {1}".format(anp, ", ".join(anps)))
    else:
        # Update anp index at template level
        template_anp_idx = anps_in_temp.index(anp)

    # If anp not at site level but exists at template level
    if anp_ref not in anps:
        op_path = "/sites/{0}/anps/-".format(site_template)
        payload.update(
            anpRef=dict(
                schemaId=schema_id,
                templateName=template,
                anpName=anp,
            ),
        )

    else:
        # Update anp index at site level
        anp_idx = anps.index(anp_ref)

    # Get EPG
    epg_ref = mso.epg_ref(schema_id=schema_id, template=template, anp=anp, epg=epg)

    # If anp exists at site level
    if "anpRef" not in payload:
        epgs = [e.get("epgRef") for e in schema_obj["sites"][site_idx]["anps"][anp_idx]["epgs"]]

    # If anp already at site level AND if epg not at site level (or) anp not at site level?
    if ("anpRef" not in payload and epg_ref not in epgs) or "anpRef" in payload:
        epgs_in_temp = [e.get("name") for e in schema_obj["templates"][template_idx]["anps"][template_anp_idx]["epgs"]]

        # If EPG not at template level - Fail
        if epg not in epgs_in_temp:
            mso.fail_json(msg="Provided EPG '{0}' does not exist. Existing EPGs: {1} epgref {2}".format(epg, ", ".join(epgs_in_temp), epg_ref))

        # EPG at template level but not at site level. Create payload at site level for EPG
        else:
            new_epg = dict(
                epgRef=dict(
                    schemaId=schema_id,
                    templateName=template,
                    anpName=anp,
                    epgName=epg,
                )
            )

            # If anp not in payload then, anp already exists at site level. New payload will only have new EPG payload
            if "anpRef" not in payload:
                op_path = "/sites/{0}/anps/{1}/epgs/-".format(site_template, anp)
                payload = new_epg
            else:
                # If anp in payload, anp exists at site level. Update payload with EPG payload
                payload["epgs"] = [new_epg]

    # Update index of EPG at site level
    else:
        epg_idx = epgs.index(epg_ref)

    if domain_association_type == "vmmDomain":
        domain_dn = "uni/vmmp-VMware/dom-{0}".format(domain_profile)
    elif domain_association_type == "l3ExtDomain":
        domain_dn = "uni/l3dom-{0}".format(domain_profile)
    elif domain_association_type == "l2ExtDomain":
        domain_dn = "uni/l2dom-{0}".format(domain_profile)
    elif domain_association_type == "physicalDomain":
        domain_dn = "uni/phys-{0}".format(domain_profile)
    elif domain_association_type == "fibreChannelDomain":
        domain_dn = "uni/fc-{0}".format(domain_profile)
    else:
        domain_dn = ""

    # Get Domains
    # If anp at site level and epg is at site level
    if "anpRef" not in payload and "epgRef" not in payload:
        domains = [dom.get("dn") for dom in schema_obj["sites"][site_idx]["anps"][anp_idx]["epgs"][epg_idx]["domainAssociations"]]
        if domain_dn in domains:
            domain_idx = domains.index(domain_dn)
            domain_path = "/sites/{0}/anps/{1}/epgs/{2}/domainAssociations/{3}".format(site_template, anp, epg, domain_idx)
            mso.existing = schema_obj["sites"][site_idx]["anps"][anp_idx]["epgs"][epg_idx]["domainAssociations"][domain_idx]

    if state == "query":
        if domain_association_type is None or domain_profile is None:
            mso.existing = schema_obj.get("sites")[site_idx]["anps"][anp_idx]["epgs"][epg_idx]["domainAssociations"]
        elif not mso.existing:
            mso.fail_json(
                msg="Domain association '{domain_association_type}/{domain_profile}' not found".format(
                    domain_association_type=domain_association_type, domain_profile=domain_profile
                )
            )
        mso.exit_json()

    domains_path = "/sites/{0}/anps/{1}/epgs/{2}/domainAssociations".format(site_template, anp, epg)
    ops = []
    new_domain = dict(
        dn=domain_dn,
        domainType=domain_association_type,
        deploymentImmediacy=deployment_immediacy,  # keeping for backworths compatibility
        deployImmediacy=deployment_immediacy,  # rename of deploymentImmediacy
        resolutionImmediacy=resolution_immediacy,
    )

    if domain_association_type == "vmmDomain":
        vmmDomainProperties = {}
        if micro_seg_vlan_type and micro_seg_vlan:
            vmmDomainProperties["microSegVlan"] = dict(vlanType=micro_seg_vlan_type, vlan=micro_seg_vlan)

        if port_encap_vlan_type and port_encap_vlan:
            vmmDomainProperties["portEncapVlan"] = dict(vlanType=port_encap_vlan_type, vlan=port_encap_vlan)

        if vlan_encap_mode:
            vmmDomainProperties["vlanEncapMode"] = vlan_encap_mode

        if allow_micro_segmentation is not None:
            vmmDomainProperties["allowMicroSegmentation"] = allow_micro_segmentation

        if switch_type:
            vmmDomainProperties["switchType"] = switch_type

        if switching_mode:
            vmmDomainProperties["switchingMode"] = switching_mode

        if enhanced_lagpolicy_name and enhanced_lagpolicy_dn:
            vmmDomainProperties["epgLagPol"] = dict(enhancedLagPol=dict(name=enhanced_lagpolicy_name, dn=enhanced_lagpolicy_dn))

        if delimiter:
            vmmDomainProperties["delimiter"] = delimiter

        if binding_type:
            vmmDomainProperties["bindingType"] = binding_type
            vmmDomainProperties["numPorts"] = num_ports

        if port_allocation:
            vmmDomainProperties["portAllocation"] = port_allocation

        if netflow_pref:
            vmmDomainProperties["netflowPref"] = netflow_pref

        if allow_promiscuous:
            vmmDomainProperties["allowPromiscuous"] = allow_promiscuous

        if forged_transmits:
            vmmDomainProperties["forgedTransmits"] = forged_transmits

        if mac_changes:
            vmmDomainProperties["macChanges"] = mac_changes

        if custom_epg_name:
            vmmDomainProperties["customEpgName"] = custom_epg_name

        if vmmDomainProperties:
            new_domain["vmmDomainProperties"] = vmmDomainProperties
            new_domain.update(vmmDomainProperties)

    # If payload is empty, anp and EPG already exist at site level
    if not payload:
        op_path = domains_path + "/-"
        payload = new_domain

    # If payload exists
    else:
        # If anp already exists at site level...(AND payload != epg as well?)
        if "anpRef" not in payload:
            payload["domainAssociations"] = [new_domain]
        else:
            payload["epgs"][0]["domainAssociations"] = [new_domain]

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing and domain_path:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=domain_path))
    elif state == "present":
        mso.sanitize(payload, collate=True)

        if mso.existing and domain_path:
            ops.append(dict(op="replace", path=domain_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=op_path, value=mso.sent))

        mso.existing = new_domain

    if not module.check_mode:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import json
import shlex
import typing as t

from ansible.module_utils.basic import AnsibleModule, env_fallback
from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.docker.plugins.module_utils._api.auth import (
    resolve_repository_name,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DEFAULT_DOCKER_HOST,
    DEFAULT_TLS,
    DEFAULT_TLS_VERIFY,
    DOCKER_MUTUALLY_EXCLUSIVE,
    DOCKER_REQUIRED_TOGETHER,
    sanitize_result,
)
from ansible_collections.community.docker.plugins.module_utils._version import (
    LooseVersion,
)

if t.TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


DOCKER_COMMON_ARGS = {
    "docker_cli": {"type": "path"},
    "docker_host": {
        "type": "str",
        "fallback": (env_fallback, ["DOCKER_HOST"]),
        "aliases": ["docker_url"],
    },
    "tls_hostname": {
        "type": "str",
        "fallback": (env_fallback, ["DOCKER_TLS_HOSTNAME"]),
    },
    "api_version": {
        "type": "str",
        "default": "auto",
        "fallback": (env_fallback, ["DOCKER_API_VERSION"]),
        "aliases": ["docker_api_version"],
    },
    "ca_path": {"type": "path", "aliases": ["ca_cert", "tls_ca_cert", "cacert_path"]},
    "client_cert": {"type": "path", "aliases": ["tls_client_cert", "cert_path"]},
    "client_key": {"type": "path", "aliases": ["tls_client_key", "key_path"]},
    "tls": {
        "type": "bool",
        "default": DEFAULT_TLS,
        "fallback": (env_fallback, ["DOCKER_TLS"]),
    },
    "validate_certs": {
        "type": "bool",
        "default": DEFAULT_TLS_VERIFY,
        "fallback": (env_fallback, ["DOCKER_TLS_VERIFY"]),
        "aliases": ["tls_verify"],
    },
    # "debug": {"type": "bool", "default: False},
    "cli_context": {"type": "str"},
}


class DockerException(Exception):
    pass


class AnsibleDockerClientBase:
    docker_api_version_str: str | None
    docker_api_version: LooseVersion | None

    def __init__(
        self,
        common_args: dict[str, t.Any],
        min_docker_api_version: str | None = None,
        needs_api_version: bool = True,
    ) -> None:
        self._environment: dict[str, str] = {}
        if common_args["tls_hostname"]:
            self._environment["DOCKER_TLS_HOSTNAME"] = common_args["tls_hostname"]
        if common_args["api_version"] and common_args["api_version"] != "auto":
            self._environment["DOCKER_API_VERSION"] = common_args["api_version"]
        cli = common_args.get("docker_cli")
        if cli is None:
            try:
                cli = get_bin_path("docker")
            except ValueError:
                self.fail(
                    "Cannot find docker CLI in path. Please provide it explicitly with the docker_cli parameter"
                )
        self._cli = cli
        self._cli_base = [self._cli]
        docker_host = common_args["docker_host"]
        if not docker_host and not common_args["cli_context"]:
            docker_host = DEFAULT_DOCKER_HOST
        if docker_host:
            self._cli_base.extend(["--host", docker_host])
        if common_args["validate_certs"]:
            self._cli_base.append("--tlsverify")
        elif common_args["tls"]:
            self._cli_base.append("--tls")
        if common_args["ca_path"]:
            self._cli_base.extend(["--tlscacert", common_args["ca_path"]])
        if common_args["client_cert"]:
            self._cli_base.extend(["--tlscert", common_args["client_cert"]])
        if common_args["client_key"]:
            self._cli_base.extend(["--tlskey", common_args["client_key"]])
        if common_args["cli_context"]:
            self._cli_base.extend(["--context", common_args["cli_context"]])

        # `--format json` was only added as a shorthand for `--format {{ json . }}` in Docker 23.0
        dummy, self._version, dummy2 = self.call_cli_json(
            "version", "--format", "{{ json . }}", check_rc=True
        )
        self._info: dict[str, t.Any] | None = None

        if needs_api_version:
            api_version_string = self._version["Server"].get(
                "ApiVersion"
            ) or self._version["Server"].get("APIVersion")
            if not isinstance(self._version.get("Server"), dict) or not isinstance(
                api_version_string, str
            ):
                self.fail(
                    "Cannot determine Docker Daemon information. Are you maybe using podman instead of docker?"
                )
            self.docker_api_version_str = to_text(api_version_string)
            self.docker_api_version = LooseVersion(self.docker_api_version_str)
            min_docker_api_version = min_docker_api_version or "1.25"
            if self.docker_api_version < LooseVersion(min_docker_api_version):
                self.fail(
                    f"Docker API version is {self.docker_api_version_str}. Minimum version required is {min_docker_api_version}."
                )
        else:
            self.docker_api_version_str = None
            self.docker_api_version = None
            if min_docker_api_version is not None:
                self.fail(
                    "Internal error: cannot have needs_api_version=False with min_docker_api_version not None"
                )

    def log(self, msg: str, pretty_print: bool = False) -> None:
        pass
        # if self.debug:
        #     from .util import log_debug
        #     log_debug(msg, pretty_print=pretty_print)

    def get_cli(self) -> str:
        return self._cli

    def get_version_info(self) -> str:
        return self._version

    def _compose_cmd(self, args: t.Sequence[str]) -> list[str]:
        return self._cli_base + list(args)

    def _compose_cmd_str(self, args: t.Sequence[str]) -> str:
        return " ".join(shlex.quote(a) for a in self._compose_cmd(args))

    @abc.abstractmethod
    def call_cli(
        self,
        *args: str,
        check_rc: bool = False,
        data: bytes | None = None,
        cwd: str | None = None,
        environ_update: dict[str, str] | None = None,
    ) -> tuple[int, bytes, bytes]:
        pass

    def call_cli_json(
        self,
        *args: str,
        check_rc: bool = False,
        data: bytes | None = None,
        cwd: str | None = None,
        environ_update: dict[str, str] | None = None,
        warn_on_stderr: bool = False,
    ) -> tuple[int, t.Any, bytes]:
        rc, stdout, stderr = self.call_cli(
            *args, check_rc=check_rc, data=data, cwd=cwd, environ_update=environ_update
        )
        if warn_on_stderr and stderr:
            self.warn(to_text(stderr))
        try:
            data = json.loads(stdout)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(
                f"Error while parsing JSON output of {self._compose_cmd_str(args)}: {exc}\nJSON output: {to_text(stdout)}\n\nError output:\n{to_text(stderr)}",
                cmd=self._compose_cmd_str(args),
                rc=rc,
                stdout=stdout,
                stderr=stderr,
            )
        return rc, data, stderr

    def call_cli_json_stream(
        self,
        *args: str,
        check_rc: bool = False,
        data: bytes | None = None,
        cwd: str | None = None,
        environ_update: dict[str, str] | None = None,
        warn_on_stderr: bool = False,
    ) -> tuple[int, list[t.Any], bytes]:
        rc, stdout, stderr = self.call_cli(
            *args, check_rc=check_rc, data=data, cwd=cwd, environ_update=environ_update
        )
        if warn_on_stderr and stderr:
            self.warn(to_text(stderr))
        result = []
        try:
            for line in stdout.splitlines():
                line = line.strip()
                if line.startswith(b"{"):
                    result.append(json.loads(line))
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(
                f"Error while parsing JSON output of {self._compose_cmd_str(args)}: {exc}\nJSON output: {to_text(stdout)}\n\nError output:\n{to_text(stderr)}",
                cmd=self._compose_cmd_str(args),
                rc=rc,
                stdout=stdout,
                stderr=stderr,
            )
        return rc, result, stderr

    @abc.abstractmethod
    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        pass

    @abc.abstractmethod
    def warn(self, msg: str) -> None:
        pass

    @abc.abstractmethod
    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
    ) -> None:
        pass

    def get_cli_info(self) -> dict[str, t.Any]:
        if self._info is None:
            dummy, self._info, dummy2 = self.call_cli_json(
                "info", "--format", "{{ json . }}", check_rc=True
            )
        return self._info

    def get_client_plugin_info(self, component: str) -> dict[str, t.Any] | None:
        cli_info = self.get_cli_info()
        if not isinstance(cli_info.get("ClientInfo"), dict):
            self.fail(
                "Cannot determine Docker client information. Are you maybe using podman instead of docker?"
            )
        for plugin in cli_info["ClientInfo"].get("Plugins") or []:
            if plugin.get("Name") == component:
                return plugin
        return None

    def _image_lookup(self, name: str, tag: str) -> list[dict[str, t.Any]]:
        """
        Including a tag in the name parameter sent to the Docker SDK for Python images method
        does not work consistently. Instead, get the result set for name and manually check
        if the tag exists.
        """
        dummy, images, dummy2 = self.call_cli_json_stream(
            "image",
            "ls",
            "--format",
            "{{ json . }}",
            "--no-trunc",
            "--filter",
            f"reference={name}",
            check_rc=True,
        )
        if tag:
            response = images
            images = []
            for image in response:
                if image.get("Tag") == tag or image.get("Digest") == tag:
                    images = [image]
                    break
        return images

    @t.overload
    def find_image(self, name: None, tag: str) -> None: ...

    @t.overload
    def find_image(self, name: str, tag: str) -> dict[str, t.Any] | None: ...

    def find_image(self, name: str | None, tag: str) -> dict[str, t.Any] | None:
        """
        Lookup an image (by name and tag) and return the inspection results.
        """
        if not name:
            return None

        self.log(f"Find image {name}:{tag}")
        images = self._image_lookup(name, tag)
        if not images:
            # In API <= 1.20 seeing 'docker.io/<name>' as the name of images pulled from docker hub
            registry, repo_name = resolve_repository_name(name)
            if registry == "docker.io":
                # If docker.io is explicitly there in name, the image
                # is not found in some cases (#41509)
                self.log(f"Check for docker.io image: {repo_name}")
                images = self._image_lookup(repo_name, tag)
                if not images and repo_name.startswith("library/"):
                    # Sometimes library/xxx images are not found
                    lookup = repo_name[len("library/") :]
                    self.log(f"Check for docker.io image: {lookup}")
                    images = self._image_lookup(lookup, tag)
                if not images:
                    # Last case for some Docker versions: if docker.io was not there,
                    # it can be that the image was not found either
                    # (https://github.com/ansible/ansible/pull/15586)
                    lookup = f"{registry}/{repo_name}"
                    self.log(f"Check for docker.io image: {lookup}")
                    images = self._image_lookup(lookup, tag)
                if not images and "/" not in repo_name:
                    # This seems to be happening with podman-docker
                    # (https://github.com/ansible-collections/community.docker/issues/291)
                    lookup = f"{registry}/library/{repo_name}"
                    self.log(f"Check for docker.io image: {lookup}")
                    images = self._image_lookup(lookup, tag)

        if len(images) > 1:
            self.fail(f"Daemon returned more than one result for {name}:{tag}")

        if len(images) == 1:
            rc, image, stderr = self.call_cli_json("image", "inspect", images[0]["ID"])
            if not image:
                self.log(f"Image {name}:{tag} not found.")
                return None
            if rc != 0:
                self.fail(f"Error inspecting image {name}:{tag} - {to_text(stderr)}")
            return image[0]

        self.log(f"Image {name}:{tag} not found.")
        return None

    @t.overload
    def find_image_by_id(
        self, image_id: None, accept_missing_image: bool = False
    ) -> None: ...

    @t.overload
    def find_image_by_id(
        self, image_id: str | None, accept_missing_image: bool = False
    ) -> dict[str, t.Any] | None: ...

    def find_image_by_id(
        self, image_id: str | None, accept_missing_image: bool = False
    ) -> dict[str, t.Any] | None:
        """
        Lookup an image (by ID) and return the inspection results.
        """
        if not image_id:
            return None

        self.log(f"Find image {image_id} (by ID)")
        rc, image, stderr = self.call_cli_json("image", "inspect", image_id)
        if not image:
            if not accept_missing_image:
                self.fail(f"Error inspecting image ID {image_id} - {to_text(stderr)}")
            self.log(f"Image {image_id} not found.")
            return None
        if rc != 0:
            self.fail(f"Error inspecting image ID {image_id} - {to_text(stderr)}")
        return image[0]


class AnsibleModuleDockerClient(AnsibleDockerClientBase):
    def __init__(
        self,
        argument_spec: dict[str, t.Any] | None = None,
        supports_check_mode: bool = False,
        mutually_exclusive: Sequence[Sequence[str]] | None = None,
        required_together: Sequence[Sequence[str]] | None = None,
        required_if: (
            Sequence[
                tuple[str, t.Any, Sequence[str]]
                | tuple[str, t.Any, Sequence[str], bool]
            ]
            | None
        ) = None,
        required_one_of: Sequence[Sequence[str]] | None = None,
        required_by: Mapping[str, Sequence[str]] | None = None,
        min_docker_api_version: str | None = None,
        fail_results: dict[str, t.Any] | None = None,
        needs_api_version: bool = True,
    ) -> None:
        # Modules can put information in here which will always be returned
        # in case client.fail() is called.
        self.fail_results = fail_results or {}

        merged_arg_spec = {}
        merged_arg_spec.update(DOCKER_COMMON_ARGS)
        if argument_spec:
            merged_arg_spec.update(argument_spec)
            self.arg_spec = merged_arg_spec

        mutually_exclusive_params: list[Sequence[str]] = [
            ("docker_host", "cli_context")
        ]
        mutually_exclusive_params += DOCKER_MUTUALLY_EXCLUSIVE
        if mutually_exclusive:
            mutually_exclusive_params += mutually_exclusive

        required_together_params: list[Sequence[str]] = []
        required_together_params += DOCKER_REQUIRED_TOGETHER
        if required_together:
            required_together_params += required_together

        self.module = AnsibleModule(
            argument_spec=merged_arg_spec,
            supports_check_mode=supports_check_mode,
            mutually_exclusive=mutually_exclusive_params,
            required_together=required_together_params,
            required_if=required_if,
            required_one_of=required_one_of,
            required_by=required_by or {},
        )

        self.debug = False  # self.module.params['debug']
        self.check_mode = self.module.check_mode
        self.diff = self.module._diff

        common_args = dict((k, self.module.params[k]) for k in DOCKER_COMMON_ARGS)
        super().__init__(
            common_args,
            min_docker_api_version=min_docker_api_version,
            needs_api_version=needs_api_version,
        )

    def call_cli(
        self,
        *args: str,
        check_rc: bool = False,
        data: bytes | None = None,
        cwd: str | None = None,
        environ_update: dict[str, str] | None = None,
    ) -> tuple[int, bytes, bytes]:
        environment = self._environment.copy()
        if environ_update:
            environment.update(environ_update)
        rc, stdout, stderr = self.module.run_command(
            self._compose_cmd(args),
            binary_data=True,
            check_rc=check_rc,
            cwd=cwd,
            data=data,
            encoding=None,
            environ_update=environment,
            expand_user_and_vars=False,
            ignore_invalid_cwd=False,
        )
        return rc, stdout, stderr

    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        self.fail_results.update(kwargs)
        self.module.fail_json(msg=msg, **sanitize_result(self.fail_results))

    def warn(self, msg: str) -> None:
        self.module.warn(msg)

    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
    ) -> None:
        self.module.deprecate(
            msg, version=version, date=date, collection_name=collection_name
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_product_facts
short_description: Provides Windows product and license information
description:
- Provides Windows product and license information.
version_added: 2.7.0
author:
- Dag Wieers (@dagwieers)
'''

EXAMPLES = r'''
- name: Get product id and product key
  ansible.windows.win_product_facts:

- name: Display Windows edition
  debug:
    var: ansible_os_license_edition

- name: Display Windows license status
  debug:
    var: ansible_os_license_status
'''

RETURN = r'''
ansible_facts:
  description: Dictionary containing all the detailed information about the Windows product and license.
  returned: always
  type: complex
  contains:
    ansible_os_license_channel:
      description: The Windows license channel.
      returned: always
      type: str
      sample: Volume:MAK
    ansible_os_license_edition:
      description: The Windows license edition.
      returned: always
      type: str
      sample: Windows(R) ServerStandard edition
    ansible_os_license_status:
      description: The Windows license status.
      returned: always
      type: str
      sample: Licensed
    ansible_os_product_id:
      description: The Windows product ID.
      returned: always
      type: str
      sample: 00326-10000-00000-AA698
    ansible_os_product_key:
      description: The Windows product key.
      returned: always
      type: str
      sample: T49TD-6VFBW-VV7HY-B2PXY-MY47H
'''



#!/usr/bin/python

# Copyright (c) 2013, Nimbis Services, Inc.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: htpasswd
short_description: Manage user files for basic authentication
description:
  - Add and remove username/password entries in a password file using htpasswd.
  - This is used by web servers such as Apache and Nginx for basic authentication.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  path:
    type: path
    required: true
    aliases: [dest, destfile]
    description:
      - Path to the file that contains the usernames and passwords.
  name:
    type: str
    required: true
    aliases: [username]
    description:
      - User name to add or remove.
  password:
    type: str
    description:
      - Password associated with user.
      - Must be specified if user does not exist yet.
  hash_scheme:
    type: str
    default: "apr_md5_crypt"
    description:
      - Hashing scheme to be used. As well as the four choices listed here, you can also use any other hash supported by passlib,
        such as V(portable_apache22) and V(host_apache24); or V(md5_crypt) and V(sha256_crypt), which are Linux passwd hashes.
        Only some schemes in addition to the four choices below are compatible with Apache or Nginx, and supported schemes
        depend on C(passlib) version and its dependencies.
      - See U(https://passlib.readthedocs.io/en/stable/lib/passlib.apache.html#passlib.apache.HtpasswdFile) parameter C(default_scheme).
      - 'Some of the available choices might be: V(apr_md5_crypt), V(des_crypt), V(ldap_sha1), V(plaintext).'
      - 'B(WARNING): The module has no mechanism to determine the O(hash_scheme) of an existing entry, therefore, it does
        not detect whether the O(hash_scheme) has changed. If you want to change the scheme, you must remove the existing
        entry and then create a new one using the new scheme.'
    aliases: [crypt_scheme]
  state:
    type: str
    choices: [present, absent]
    default: "present"
    description:
      - Whether the user entry should be present or not.
  create:
    type: bool
    default: true
    description:
      - Used with O(state=present). If V(true), the file is created if it does not exist. Conversely, if set to V(false) and
        the file does not exist, it fails.
notes:
  - This module depends on the C(passlib) Python library, which needs to be installed on all target systems.
  - 'On Debian < 11, Ubuntu <= 20.04, or Fedora: install C(python-passlib).'
  - 'On Debian, Ubuntu: install C(python3-passlib).'
  - 'On RHEL or CentOS: Enable EPEL, then install C(python-passlib).'
  - To use V(bcrypt) as O(hash_scheme), the C(bcrypt) Python library must also be installed.
    Due to incompatibilities in C(passlib) 1.7.x, use C(bcrypt<4.2).
requirements: [passlib>=1.6]
author: "Ansible Core Team"
extends_documentation_fragment:
  - ansible.builtin.files
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Add a user to a password file and ensure permissions are set
  community.general.htpasswd:
    path: /etc/nginx/passwdfile
    name: janedoe
    password: '9s36?;fyNp'
    owner: root
    group: www-data
    mode: '0640'

- name: Remove a user from a password file
  community.general.htpasswd:
    path: /etc/apache2/passwdfile
    name: foobar
    state: absent

- name: Add a user to a password file suitable for use by libpam-pwdfile
  community.general.htpasswd:
    path: /etc/mail/passwords
    name: alex
    password: oedu2eGh
    hash_scheme: md5_crypt
"""


import os
import tempfile

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils import _deps as deps

with deps.declare("passlib"):
    # Apparently the type infos don't know htpasswd_context, which *does* exist
    # (but isn't mentioned in the documentation for some reason)
    from passlib.apache import HtpasswdFile, htpasswd_context  # type: ignore[attr-defined]
    from passlib.context import CryptContext  # type: ignore[assignment]


apache_hashes = ["apr_md5_crypt", "des_crypt", "ldap_sha1", "plaintext"]


def create_missing_directories(dest):
    destpath = os.path.dirname(dest)
    if not os.path.exists(destpath):
        os.makedirs(destpath)


def obtain_crypt_context(hash_scheme):
    if hash_scheme in apache_hashes or hash_scheme in htpasswd_context.schemes():
        # Use htpasswd_context for all officially-supported schemes, including bcrypt
        # (htpasswd_context produces Apache-compatible $2y$ bcrypt, not $2b$)
        return htpasswd_context
    try:
        return CryptContext(schemes=[hash_scheme] + apache_hashes)
    except KeyError:
        # hash_scheme is a HtpasswdFile alias (e.g. portable, portable_apache_24)
        # that is not a valid passlib scheme name; let HtpasswdFile resolve it natively
        return htpasswd_context


def present(dest, username, password, hash_scheme, create, check_mode):
    """Ensures user is present

    Returns (msg, changed)"""
    context = obtain_crypt_context(hash_scheme)
    if not os.path.exists(dest):
        if not create:
            raise ValueError(f"Destination {dest} does not exist")
        if check_mode:
            return (f"Create {dest}", True)
        create_missing_directories(dest)
        ht = HtpasswdFile(dest, new=True, default_scheme=hash_scheme, context=context)
        ht.set_password(username, password)
        ht.save()
        return (f"Created {dest} and added {username}", True)
    else:
        ht = HtpasswdFile(dest, new=False, default_scheme=hash_scheme, context=context)

        found = ht.check_password(username, password)

        if found:
            return (f"{username} already present", False)
        else:
            if not check_mode:
                ht.set_password(username, password)
                ht.save()
            return (f"Add/update {username}", True)


def absent(dest, username, check_mode):
    """Ensures user is absent

    Returns (msg, changed)"""
    ht = HtpasswdFile(dest, new=False)

    if username not in ht.users():
        return (f"{username} not present", False)
    else:
        if not check_mode:
            ht.delete(username)
            ht.save()
        return (f"Remove {username}", True)


def check_file_attrs(module, changed, message):
    file_args = module.load_file_common_arguments(module.params)
    if module.set_fs_attributes_if_different(file_args, False):
        if changed:
            message += " and "
        changed = True
        message += "ownership, perms or SE linux context changed"

    return message, changed


def main():
    arg_spec = dict(
        path=dict(type="path", required=True, aliases=["dest", "destfile"]),
        name=dict(type="str", required=True, aliases=["username"]),
        password=dict(type="str", no_log=True),
        hash_scheme=dict(type="str", default="apr_md5_crypt", aliases=["crypt_scheme"]),
        state=dict(type="str", default="present", choices=["present", "absent"]),
        create=dict(type="bool", default=True),
    )
    module = AnsibleModule(argument_spec=arg_spec, add_file_common_args=True, supports_check_mode=True)

    path = module.params["path"]
    username = module.params["name"]
    password = module.params["password"]
    hash_scheme = module.params["hash_scheme"]
    state = module.params["state"]
    create = module.params["create"]
    check_mode = module.check_mode

    deps.validate(module)

    # TODO double check if this hack below is still needed.
    # Check file for blank lines in effort to avoid "need more than 1 value to unpack" error.
    try:
        with open(path) as f:
            lines = f.readlines()

        # If the file gets edited, it returns true, so only edit the file if it has blank lines
        strip = False
        for line in lines:
            if not line.strip():
                strip = True
                break

        if strip:
            # If check mode, create a temporary file
            if check_mode:
                temp = tempfile.NamedTemporaryFile()
                path = temp.name
            with open(path, "w") as f:
                f.writelines(line for line in lines if line.strip())

    except OSError:
        # No preexisting file to remove blank lines from
        pass

    try:
        if state == "present":
            (msg, changed) = present(path, username, password, hash_scheme, create, check_mode)
        elif state == "absent":
            if not os.path.exists(path):
                module.warn(f"{path} does not exist")
                module.exit_json(msg=f"{username} not present", changed=False)
            (msg, changed) = absent(path, username, check_mode)
        else:
            module.fail_json(msg=f"Invalid state: {state}")
            return  # needed to make pylint happy

        (msg, changed) = check_file_attrs(module, changed, msg)
        module.exit_json(msg=msg, changed=changed)
    except Exception as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



# Copyright (c) 2016, Samuel Boucher <boucher.samuel.c@gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: keyring
author:
  - Samuel Boucher (!UNKNOWN) <boucher.samuel.c@gmail.com>
requirements:
  - keyring (python library)
short_description: Grab secrets from the OS keyring
description:
  - Allows you to access data stored in the OS provided keyring/keychain.
options:
  _terms:
    description:
      - List of pairs of service and user name.
      - Every entry must be of the form V(servicename username).
    type: list
    elements: str
"""

EXAMPLES = r"""
- name: output secrets to screen (BAD IDEA)
  ansible.builtin.debug:
    msg: "Password: {{item}}"
  with_community.general.keyring:
    - 'servicename username'

- name: access mysql with password from keyring
  community.mysql.mysql_db:
    login_password: "{{ lookup('community.general.keyring', 'mysql joe') }}"
    login_user: joe
"""

RETURN = r"""
_raw:
  description: Secrets stored.
  type: list
  elements: str
"""

HAS_KEYRING = True

from ansible.errors import AnsibleError
from ansible.utils.display import Display

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms

try:
    import keyring
except ImportError:
    HAS_KEYRING = False

from ansible.plugins.lookup import LookupBase

display = Display()


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        if not HAS_KEYRING:
            raise AnsibleError("Can't LOOKUP(keyring): missing required python library 'keyring'")

        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)

        display.vvvv(f"keyring: {keyring.get_keyring()}")
        ret = []
        for term in terms:
            (servicename, username) = (term.split()[0], term.split()[1])
            display.vvvv(f"username: {username}, servicename: {servicename} ")
            password = keyring.get_password(servicename, username)
            if password is None:
                raise AnsibleError(f"servicename: {servicename} for user {username} not found")
            ret.append(password.rstrip())
        return ret



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
filter plugin file for ipaddr filters: cidr_merge
"""
from __future__ import absolute_import, division, print_function

from functools import partial

from ansible.errors import AnsibleFilterError

from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)
from ansible_collections.ansible.utils.plugins.plugin_utils.base.ipaddr_utils import _need_netaddr


__metaclass__ = type


try:
    from jinja2.filters import pass_environment
except ImportError:
    from jinja2.filters import environmentfilter as pass_environment

try:
    import netaddr

    HAS_NETADDR = True
except ImportError:
    # in this case, we'll make the filters return error messages (see bottom)
    HAS_NETADDR = False
else:

    class mac_linux(netaddr.mac_unix):
        pass

    mac_linux.word_fmt = "%.2x"

DOCUMENTATION = """
    name: ipmath
    author: Ashwini Mhatre (@amhatre)
    version_added: "2.5.0"
    short_description: This filter is designed to do simple IP math/arithmetic.
    description:
        - This filter is designed to do simple IP math/arithmetic.
    options:
        value:
            description:
            - list of subnets or individual address or any other values input for ipaddr plugin
            type: str
            required: True
        amount:
            type: int
            description: integer for arithmetic. Example -1,2,3
"""

EXAMPLES = r"""
#### examples
# Ipmath filter plugin with different arthmetic.
# Get the next fifth address based on an IP address
- debug:
    msg: "{{ '192.168.1.5' | ansible.utils.ipmath(5) }}"

# Get the tenth previous address based on an IP address
- debug:
    msg: "{{ '192.168.1.5' | ansible.utils.ipmath(-10) }}"

# Get the next fifth address using CIDR notation
- debug:
    msg: "{{ '192.168.1.1/24' | ansible.utils.ipmath(5) }}"

# Get the previous fifth address using CIDR notation
- debug:
    msg: "{{ '192.168.1.6/24' | ansible.utils.ipmath(-5) }}"

# Get the previous tenth address using cidr notation
# It returns a address of the previous network range
- debug:
    msg: "{{ '192.168.2.6/24' | ansible.utils.ipmath(-10) }}"

# Get the next tenth address in IPv6
- debug:
    msg: "{{ '2001::1' | ansible.utils.ipmath(10) }}"

# Get the previous tenth address in IPv6
- debug:
    msg: "{{ '2001::5' | ansible.utils.ipmath(-10) }}"

# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "192.168.1.10"
# }
#
# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "192.168.0.251"
# }
#
# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "192.168.1.6"
# }
#
# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "192.168.1.1"
# }
#
# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "192.168.1.252"
# }
#
# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "2001::b"
# }
#
# TASK [debug] **********************************************************************************************************
# ok: [localhost] => {
#     "msg": "2000:ffff:ffff:ffff:ffff:ffff:ffff:fffb"
# }
"""

RETURN = """
  data:
    type: str
    description:
      - Returns result of IP math/arithmetic.
"""


@pass_environment
def _ipmath(*args, **kwargs):
    """Convert the given data from json to xml."""
    keys = ["value", "amount"]
    data = dict(zip(keys, args[1:]))
    data.update(kwargs)
    aav = AnsibleArgSpecValidator(data=data, schema=DOCUMENTATION, name="ipmath")
    valid, errors, updated_data = aav.validate()
    if not valid:
        raise AnsibleFilterError(errors)
    return ipmath(**updated_data)


def ipmath(value, amount):
    try:
        if "/" in value:
            ip = netaddr.IPNetwork(value).ip
        else:
            ip = netaddr.IPAddress(value)
    except (netaddr.AddrFormatError, ValueError):
        msg = "You must pass a valid IP address; {0} is invalid".format(value)
        raise AnsibleFilterError(msg)

    if not isinstance(amount, int):
        msg = ("You must pass an integer for arithmetic; " "{0} is not a valid integer").format(
            amount,
        )
        raise AnsibleFilterError(msg)

    return str(ip + amount)


class FilterModule(object):
    """IP address and network manipulation filters"""

    filter_map = {
        # This filter is designed to do simple IP math/arithmetic
        "ipmath": _ipmath,
    }

    def filters(self):
        """ipmath filter"""
        if HAS_NETADDR:
            return self.filter_map
        else:
            return dict((f, partial(_need_netaddr, f)) for f in self.filter_map)



#!/usr/bin/python

# Copyright (c) 2026, Masaru Onodera (@masa-orca)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_project_approvals
short_description: Manage project-level merge request approvals settings on GitLab Server
version_added: 13.1.0
description:
  - This module allows to manage project-level merge request approval configurations.
author:
  - "Masaru Onodera (@masa-orca)"
requirements:
  - python-gitlab python module
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  project:
    description:
      - The name or full path of the GitLab project.
    required: true
    type: str

  approvals_before_merge:
    description:
      - The number of approvals required before a merge request can be merged.
      - Deprecated in GitLab 12.3.
    type: int

  reset_approvals_on_push:
    description:
      - Reset approvals on new commit.
    type: bool

  disable_overriding_approvers_per_merge_request:
    description:
      - Disable overriding approvers per merge request.
    type: bool

  merge_requests_author_approval:
    description:
      - Allow authors to self-approve merge requests.
    type: bool

  merge_requests_disable_committers_approval:
    description:
      - Prevent committers from approving merge requests.
    type: bool

  selective_code_owner_removals:
    description:
      - Reset approvals from Code Owners if their files are changed.
    type: bool

  require_password_to_approve:
    description:
      - Require password reauthentication to approve a merge request.
      - Deprecated in GitLab 16.9.
    type: bool

  require_reauthentication_to_approve:
    description:
      - Require user reauthentication to approve a merge request.
    type: bool
"""

EXAMPLES = r"""
- name: Configure GitLab Project approval settings
  community.general.gitlab_project_approvals:
    api_url: https://gitlab.example.com/
    api_token: "{{ api_token }}"
    project: my_group/my_project
    reset_approvals_on_push: true
    merge_requests_author_approval: false
    merge_requests_disable_committers_approval: true
"""

RETURN = r"""
project_approvals:
  description: The updated GitLab project approval settings.
  returned: always
  type: dict
  sample:
    approvals_before_merge: 2
    reset_approvals_on_push: true
    disable_overriding_approvers_per_merge_request: false
    merge_requests_author_approval: false
    merge_requests_disable_committers_approval: true
"""


from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    find_project,
    gitlab_authentication,
)


class GitlabProjectApprovals:
    def __init__(self, module, gitlab_instance):
        self._module = module
        self._gitlab = gitlab_instance
        self.project = self.get_project(module.params["project"])

    def get_project(self, project_name):
        project = find_project(self._gitlab, project_name)
        if project is None:
            self._module.fail_json(msg=f"Project {project_name} not found or insufficient permissions.")
        return project

    def update_approval_settings(self, options):
        changed = False
        try:
            approval_settings = self.project.approvals.get()
        except Exception as e:
            self._module.fail_json(msg=f"Failed to get project approval settings: {e}")

        # Check which options differ from current settings and apply them
        for key, value in options.items():
            if value is not None:
                current_value = getattr(approval_settings, key, None)
                if current_value != value:
                    setattr(approval_settings, key, value)
                    changed = True

        if changed:
            if not self._module.check_mode:
                try:
                    approval_settings.save()
                except Exception as e:
                    self._module.fail_json(msg=f"Failed to update project approval settings: {e}")

        return changed, approval_settings.attributes


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        dict(
            project=dict(type="str", required=True),
            approvals_before_merge=dict(type="int"),
            reset_approvals_on_push=dict(type="bool"),
            disable_overriding_approvers_per_merge_request=dict(type="bool"),
            merge_requests_author_approval=dict(type="bool"),
            merge_requests_disable_committers_approval=dict(type="bool"),
            selective_code_owner_removals=dict(type="bool"),
            require_password_to_approve=dict(type="bool"),
            require_reauthentication_to_approve=dict(type="bool"),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
        ],
        required_together=[
            ["api_username", "api_password"],
        ],
        required_one_of=[["api_username", "api_token", "api_oauth_token", "api_job_token"]],
        supports_check_mode=True,
    )

    gitlab_instance = gitlab_authentication(module)

    gitlab_project_approvals = GitlabProjectApprovals(module, gitlab_instance)

    options = {
        "approvals_before_merge": module.params["approvals_before_merge"],
        "reset_approvals_on_push": module.params["reset_approvals_on_push"],
        "disable_overriding_approvers_per_merge_request": module.params[
            "disable_overriding_approvers_per_merge_request"
        ],
        "merge_requests_author_approval": module.params["merge_requests_author_approval"],
        "merge_requests_disable_committers_approval": module.params["merge_requests_disable_committers_approval"],
        "selective_code_owner_removals": module.params["selective_code_owner_removals"],
        "require_password_to_approve": module.params["require_password_to_approve"],
        "require_reauthentication_to_approve": module.params["require_reauthentication_to_approve"],
    }

    changed, approval_attrs = gitlab_project_approvals.update_approval_settings(options)

    module.exit_json(changed=changed, project_approvals=approval_attrs)


if __name__ == "__main__":
    main()
